"""Evaluation answer."""

import asyncio
import logging
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Sequence, Tuple, Type

from dbgpt.core import HumanPromptTemplate, ModelMessage
from dbgpt.core.awel import IteratorTrigger, JoinOperator, MapOperator
from dbgpt.core.interface.evaluation import (
    DatasetType,
    EvaluationMetric,
    EvaluationResult,
    Evaluator,
)
from dbgpt.core.interface.llm import LLMClient, ModelRequest

logger = logging.getLogger(__name__)

ANSWER_RELEVANCY_EVALUATE_PROMPT_TEMPLATE = """
你是一个智能答疑专家, 你的任务是根据用户的问题和已经相关的文档给问答的答案进行严格的\
打分.
你将会得到以下输入信息:
- 用户的问题
- 答案的知识来源，用于评估中的参考.
- 问答系统生成的答案

约束条件:
- 你的工作是判断生成回答的相关性和正确性，并且输出一个代表整体评价的单一得分。
- 你所返回的答案必须包含一个得分，且只能包含得分。
- 不能返回任何其他格式的答案。
在另外一行中需要提供你对得分的理由。

遵循以下评分指南：
你的得分必须在0到5之间，其中0是最差的，5是最好的。
如果生成的答案包含LLM ERROR信息，你应该给出0分。
如果生成的答案与相关的参考内容基本不相关，你应该给出1分。
如果生成的答案与相关的参考内容有点相关，但回答并没有十分的详细，你应该给出2分。
如果生成的答案与相关的参考内容非常相关，但回答并没有十分的详细，你应该给出3分。
如果生成的答案与相关的参考内容相关且完全正确，并且十分详细地回答用户的问题，你应该给出4分。
如果生成的答案与相关的参考内容相关且完全正确，并且十分详细地回答用户的问题，还有自己的建议和思考，你应该给出5分。

用户的问题是:
{query}
相关的参考:
{context}
模型生成的答案:
{answer}

Example Response:
4.0
如果生成的答案与相关的参考内容相关且完全正确，并且十分详细地回答用户的问题，你应该给出4分。

"""


class LLMEvaluationMetric(EvaluationMetric):
    """LLM Relevancy metric."""

    prompt_template = ANSWER_RELEVANCY_EVALUATE_PROMPT_TEMPLATE

    def __init__(
        self,
        llm_client: LLMClient,
        model_name: Optional[str] = None,
        prompt_template: Optional[str] = None,
    ):
        """Create a SimilarityMetric with embeddings."""
        self._llm_client = llm_client
        self._model_name = model_name
        self._prompt_template = prompt_template

    async def compute(
        self,
        prediction: str,
        contexts: Optional[Sequence[str]] = None,
        query: Optional[str] = None,
    ) -> EvaluationResult:
        """Compute the evaluation metric.

        Args:
            prediction(List[str]): The retrieved chunks from the retriever.
            contexts(Sequence[str]): The contexts from dataset.
            query:(Optional[str]) The query text.

        Returns:
            BaseEvaluationResult: The evaluation result.
                The score is the mean of the cosine similarity between the prediction
                and the contexts.
        """
        if not prediction or not contexts:
            return EvaluationResult(
                query=query,
                prediction=prediction,
                contexts=contexts,
                score=0.0,
            )
        evaluate_result = await self._generate_llm_result(query, prediction, contexts)
        score, feedback = self._parse_evaluate_response(evaluate_result)
        return EvaluationResult(
            query=query,
            prediction=prediction,
            contexts=contexts,
            score=score,
            feedback=feedback,
        )

    async def _generate_llm_result(
        self,
        query: str | None = None,
        answer: str | None = None,
        contexts: Optional[Sequence[str]] = None,
        **kwargs: Any,
    ) -> str:
        """Generate llm result."""
        template = HumanPromptTemplate.from_template(self._prompt_template)
        messages = template.format_messages(
            query=query, context=contexts, answer=answer
        )
        if not self._model_name:
            models = await self._llm_client.models()
            if not models:
                raise Exception("No models available")
            self._model_name = models[0].model
            logger.info(f"Using model {self._model_name} to evaluate")

        model_messages = ModelMessage.from_base_messages(messages)
        request = ModelRequest(model=self._model_name, messages=model_messages)
        response = await self._llm_client.generate(request=request)

        if not response.success:
            code = str(response.error_code)
            reason = response.text
            logger.error(f"request llm failed ({code}) {reason}")
            return f"request llm failed ({code}) {reason}"
        return response.text

    @abstractmethod
    def _parse_evaluate_response(
        self, response: str, limit: Optional[int] = None
    ) -> Tuple[Any, Any]:
        """Parse llm evaluate response."""


class AnswerRelevancyMetric(LLMEvaluationMetric):
    """Answer Relevancy metric."""

    prompt_template = ANSWER_RELEVANCY_EVALUATE_PROMPT_TEMPLATE

    def __init__(
        self,
        llm_client: LLMClient,
        model_name: Optional[str] = None,
        prompt_template: Optional[str] = None,
    ):
        """Create a AnswerRelevancyMetric with llm_client.

        Args:
            llm_client(LLMClient): The llm client to use for evaluation.
            model_name(str): The model name to use for evaluation.
            prompt_template(Optional[str]): The prompt template to use for evaluation.
        """
        self._llm_client = llm_client
        self._model_name = model_name
        self._prompt_template = (
            prompt_template or ANSWER_RELEVANCY_EVALUATE_PROMPT_TEMPLATE
        )
        super().__init__(self._llm_client, self._model_name, self._prompt_template)

    def _parse_evaluate_response(
        self, text: str, limit: Optional[int] = None
    ) -> Tuple[Any, Any]:
        parts = text.split("\\n", 1)
        score = parts[0].strip()
        try:
            score = float(score)  # type: ignore # noqa
        except ValueError as e:
            logger.error(f"Error converting \\n score to float: {e}")
            parts = text.split("\n", 1)
            score = parts[0].strip()
            try:
                score = float(score)  # type: ignore # noqa
            except ValueError as e:
                logger.error(f"Error converting score to float: {e}")
                return 0.0, "parse score float error."
        feedback = parts[1].strip() if len(parts) > 1 else None
        return score, feedback


class LLMAnswerEvaluator(Evaluator):
    """Evaluator for LLM Answer relevancy."""

    def __init__(
        self,
        operator_cls: Type[MapOperator],
        llm_client: Optional[LLMClient] = None,
        operator_kwargs: Optional[Dict] = None,
    ):
        """Create a new LLMAnswerEvaluator."""
        if not operator_kwargs:
            operator_kwargs = {}
        self._operator_cls = operator_cls
        self._operator_kwargs: Dict[str, Any] = operator_kwargs
        super().__init__(llm_client=llm_client)

    async def evaluate(
        self,
        dataset: DatasetType,
        metrics: Optional[List[EvaluationMetric]] = None,
        query_key: str = "query",
        contexts_key: str = "contexts",
        prediction_key: str = "prediction",
        parallel_num: int = 1,
        **kwargs,
    ) -> List[List[EvaluationResult]]:
        """Evaluate the dataset."""
        from dbgpt.core.awel import DAG, MapOperator

        if not metrics:
            if not self.llm_client:
                raise ValueError("llm_client are required for AnswerRelevancyMetric")
            metrics = [AnswerRelevancyMetric(self.llm_client)]

        with DAG("relevancy_evaluation_dag"):
            input_task = IteratorTrigger(dataset)
            query_task: MapOperator = MapOperator(lambda x: x[query_key])
            answer_output_task = self._operator_cls(**self._operator_kwargs)
            answer_eva_task = AnswerEvaluatorOperator(
                evaluation_metrics=metrics,
                llm_client=self.llm_client,  # type: ignore # noqa
            )
            input_task >> query_task
            query_task >> answer_eva_task
            query_task >> answer_output_task >> answer_eva_task
            input_task >> MapOperator(lambda x: x[contexts_key]) >> answer_eva_task
            input_task >> answer_eva_task

        results = await input_task.trigger(parallel_num=parallel_num)
        return [item for _, item in results]


class AnswerEvaluatorOperator(JoinOperator[List[EvaluationResult]]):
    """Evaluator for Answer."""

    def __init__(
        self,
        evaluation_metrics: List[LLMEvaluationMetric],
        llm_client: Optional[LLMClient] = None,
        **kwargs,
    ):
        """Create a new AnswerEvaluatorOperator.

        Args:
            evaluation_metrics(List[LLMEvaluationMetric]): The evaluation metrics.
            llm_client(Optional[LLMClient]): The llm client to use for evaluation.
        """
        self.llm_client = llm_client
        self.evaluation_metrics = evaluation_metrics
        super().__init__(combine_function=self._do_evaluation, **kwargs)

    async def _do_evaluation(
        self,
        query: str,
        prediction: List[str],
        contexts: List[str],
        raw_dataset: Any = None,
    ) -> List[EvaluationResult]:
        """Run evaluation.

        Args:
            query(str): The query string.
            prediction(List[str]): The retrieved chunks from the retriever.
            contexts(List[str]): The contexts from dataset.
            raw_dataset(Any): The raw data(single row) from dataset.
        """
        if isinstance(contexts, str):
            contexts = [contexts]
        tasks = []
        for metric in self.evaluation_metrics:
            tasks.append(metric.compute(query, prediction, contexts))  # type: ignore # noqa
        task_results = await asyncio.gather(*tasks)
        results = []
        for result, metric in zip(task_results, self.evaluation_metrics):
            results.append(
                EvaluationResult(
                    query=query,
                    prediction=result.prediction,
                    score=result.score,
                    contexts=contexts,
                    passing=result.passing,
                    raw_dataset=raw_dataset,
                    metric_name=metric.name(),
                    feedback=result.feedback,
                )
            )
        return results



from typing import Dict, Union

import pytest

from dbgpt.core.interface.storage import (
    InMemoryStorage,
    QuerySpec,
    ResourceIdentifier,
    StorageError,
    StorageItem,
)
from dbgpt.util.serialization.json_serialization import JsonSerializer


class MockResourceIdentifier(ResourceIdentifier):
    def __init__(self, identifier: str):
        self._identifier = identifier

    @property
    def str_identifier(self) -> str:
        return self._identifier

    def to_dict(self) -> Dict:
        return {"identifier": self._identifier}


class MockStorageItem(StorageItem):
    def merge(self, other: "StorageItem") -> None:
        if not isinstance(other, MockStorageItem):
            raise ValueError("other must be a MockStorageItem")
        self.data = other.data

    def __init__(self, identifier: Union[str, MockResourceIdentifier], data):
        self._identifier_str = (
            identifier if isinstance(identifier, str) else identifier.str_identifier
        )
        self.data = data

    def to_dict(self) -> Dict:
        return {"identifier": self._identifier_str, "data": self.data}

    @property
    def identifier(self) -> ResourceIdentifier:
        return MockResourceIdentifier(self._identifier_str)


@pytest.fixture
def serializer():
    return JsonSerializer()


@pytest.fixture
def in_memory_storage(serializer):
    return InMemoryStorage(serializer)


def test_save_and_load(in_memory_storage):
    resource_id = MockResourceIdentifier("1")
    item = MockStorageItem(resource_id, "test_data")

    in_memory_storage.save(item)

    loaded_item = in_memory_storage.load(resource_id, MockStorageItem)
    assert loaded_item.data == "test_data"


def test_duplicate_save(in_memory_storage):
    item = MockStorageItem("1", "test_data")

    in_memory_storage.save(item)

    # Should raise StorageError when saving the same data
    with pytest.raises(StorageError):
        in_memory_storage.save(item)


def test_delete(in_memory_storage):
    resource_id = MockResourceIdentifier("1")
    item = MockStorageItem(resource_id, "test_data")

    in_memory_storage.save(item)
    in_memory_storage.delete(resource_id)
    # Storage should not contain the data after deletion
    assert in_memory_storage.load(resource_id, MockStorageItem) is None


def test_query(in_memory_storage):
    resource_id1 = MockResourceIdentifier("1")
    item1 = MockStorageItem(resource_id1, "test_data1")

    resource_id2 = MockResourceIdentifier("2")
    item2 = MockStorageItem(resource_id2, "test_data2")

    in_memory_storage.save(item1)
    in_memory_storage.save(item2)

    query_spec = QuerySpec(conditions={"data": "test_data1"})
    results = in_memory_storage.query(query_spec, MockStorageItem)
    assert len(results) == 1
    assert results[0].data == "test_data1"


def test_count(in_memory_storage):
    item1 = MockStorageItem("1", "test_data1")

    item2 = MockStorageItem("2", "test_data2")

    in_memory_storage.save(item1)
    in_memory_storage.save(item2)

    query_spec = QuerySpec(conditions={})
    count = in_memory_storage.count(query_spec, MockStorageItem)
    assert count == 2


def test_paginate_query(in_memory_storage):
    for i in range(10):
        resource_id = MockResourceIdentifier(str(i))
        item = MockStorageItem(resource_id, f"test_data{i}")
        in_memory_storage.save(item)

    page_size = 3
    query_spec = QuerySpec(conditions={})
    page_result = in_memory_storage.paginate_query(
        2, page_size, MockStorageItem, query_spec
    )

    assert len(page_result.items) == page_size
    assert page_result.total_count == 10
    assert page_result.total_pages == 4
    assert page_result.page == 2



from typing import List

import pytest

from .. import (
    DAG,
    BranchOperator,
    DAGContext,
    InputOperator,
    JoinOperator,
    MapOperator,
    ReduceStreamOperator,
    SimpleInputSource,
    TaskState,
    WorkflowRunner,
)
from .conftest import (
    _is_async_iterator,
)


@pytest.mark.asyncio
async def test_input_node(runner: WorkflowRunner):
    input_node = InputOperator(SimpleInputSource("hello"), task_id="112232")
    res: DAGContext[str] = await runner.execute_workflow(input_node)
    assert res.current_task_context.current_state == TaskState.SUCCESS
    assert res.current_task_context.task_output.output == "hello"

    async def new_steam_iter(n: int):
        for i in range(n):
            yield i

    num_iter = 10
    steam_input_node = InputOperator(
        SimpleInputSource(new_steam_iter(num_iter)), task_id="112232"
    )
    res: DAGContext[str] = await runner.execute_workflow(steam_input_node)
    assert res.current_task_context.current_state == TaskState.SUCCESS
    output_steam = res.current_task_context.task_output.output_stream
    assert output_steam
    assert _is_async_iterator(output_steam)
    i = 0
    async for x in output_steam:
        assert x == i
        i += 1


@pytest.mark.asyncio
async def test_map_node(runner: WorkflowRunner, stream_input_node: InputOperator):
    with DAG("test_map") as _dag:
        map_node = MapOperator(lambda x: x * 2)
        stream_input_node >> map_node
        res: DAGContext[int] = await runner.execute_workflow(map_node)
        output_steam = res.current_task_context.task_output.output_stream
        assert output_steam
        i = 0
        async for x in output_steam:
            assert x == i * 2
            i += 1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "stream_input_node, expect_sum",
    [
        ({"output_streams": [[0, 1, 2, 3]]}, 6),
        ({"output_streams": [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]]}, 55),
    ],
    indirect=["stream_input_node"],
)
async def test_reduce_node(
    runner: WorkflowRunner, stream_input_node: InputOperator, expect_sum: int
):
    with DAG("test_reduce_node") as _dag:
        reduce_node = ReduceStreamOperator(lambda x, y: x + y)
        stream_input_node >> reduce_node
        res: DAGContext[int] = await runner.execute_workflow(reduce_node)
        assert res.current_task_context.current_state == TaskState.SUCCESS
        assert not res.current_task_context.task_output.is_stream
        assert res.current_task_context.task_output.output == expect_sum


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "input_nodes",
    [
        ({"outputs": [0, 1, 2]}),
    ],
    indirect=["input_nodes"],
)
async def test_join_node(runner: WorkflowRunner, input_nodes: List[InputOperator]):
    def join_func(p1, p2, p3) -> int:
        return p1 + p2 + p3

    with DAG("test_join_node") as _dag:
        join_node = JoinOperator(join_func)
        for input_node in input_nodes:
            input_node >> join_node
        res: DAGContext[int] = await runner.execute_workflow(join_node)
        assert res.current_task_context.current_state == TaskState.SUCCESS
        assert not res.current_task_context.task_output.is_stream
        assert res.current_task_context.task_output.output == 3


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "input_node, is_odd",
    [
        ({"outputs": [0]}, False),
        ({"outputs": [1]}, True),
    ],
    indirect=["input_node"],
)
async def test_branch_node(
    runner: WorkflowRunner, input_node: InputOperator, is_odd: bool
):
    def join_func(o1, o2) -> int:
        print(f"join func result, o1: {o1}, o2: {o2}")
        return o1 or o2

    with DAG("test_join_node") as _dag:
        odd_node = MapOperator(
            lambda x: 999, task_id="odd_node", task_name="odd_node_name"
        )
        even_node = MapOperator(
            lambda x: 888, task_id="even_node", task_name="even_node_name"
        )
        join_node = JoinOperator(join_func, can_skip_in_branch=False)
        branch_node = BranchOperator(
            {lambda x: x % 2 == 1: odd_node, lambda x: x % 2 == 0: even_node}
        )
        branch_node >> odd_node >> join_node
        branch_node >> even_node >> join_node

        input_node >> branch_node

        res: DAGContext[int] = await runner.execute_workflow(join_node)
        assert res.current_task_context.current_state == TaskState.SUCCESS
        expect_res = 999 if is_odd else 888
        assert res.current_task_context.task_output.output == expect_res



"""Models Configuration Parameters."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

from dbgpt.core.interface.parameter import (
    BaseServerParameters,
    EmbeddingDeployModelParameters,
    LLMDeployModelParameters,
    RerankerDeployModelParameters,
)
from dbgpt.datasource.parameter import BaseDatasourceParameters
from dbgpt.util.configure.manager import RegisterParameters
from dbgpt.util.i18n_utils import _
from dbgpt.util.parameter_utils import BaseParameters


class WorkerType(str, Enum):
    LLM = "llm"
    TEXT2VEC = "text2vec"
    RERANKER = "reranker"

    @staticmethod
    def values():
        return [item.value for item in WorkerType]

    @staticmethod
    def to_worker_key(worker_name, worker_type: Union[str, "WorkerType"]) -> str:
        """Generate worker key from worker name and worker type

        Args:
            worker_name (str): Worker name(eg., chatglm2-6b)
            worker_type (Union[str, "WorkerType"]):
                Worker type(eg., 'llm', or [`WorkerType.LLM`])

        Returns:
            str: Generated worker key
        """
        if "@" in worker_name:
            raise ValueError(f"Invaild symbol '@' in your worker name {worker_name}")
        if isinstance(worker_type, WorkerType):
            worker_type = worker_type.value
        return f"{worker_name}@{worker_type}"

    @staticmethod
    def parse_worker_key(worker_key: str) -> Tuple[str, str]:
        """Parse worker name and worker type from worker key

        Args:
            worker_key (str): Worker key generated by [`WorkerType.to_worker_key`]

        Returns:
            Tuple[str, str]: Worker name and worker type
        """
        return tuple(worker_key.split("@"))

    @classmethod
    def from_str(cls, value: str) -> "WorkerType":
        """Convert a string to an Enum value."""
        try:
            return cls(value)
        except ValueError:
            raise ValueError(
                f"Invalid value '{value}' for {cls.__name__}. "
                f"Valid values are {cls.values()}"
            )


@dataclass
class BaseModelRegistryParameters(BaseParameters, RegisterParameters):
    """Base model registry parameters."""

    __type__ = "___model_registry_placeholder___"

    __cfg_type__ = "service"

    @classmethod
    def _from_dict_(
        cls, data: Dict, prepare_data_func, converter
    ) -> Optional["BaseModelRegistryParameters"]:
        db = data.get("database", None)
        if db:
            real_data = prepare_data_func(BaseDatasourceParameters, data["database"])
            real_data["type"] = data["database"]["type"]
            database = converter(real_data, BaseDatasourceParameters)
            return DBModelRegistryParameters(database=database)
        return None


@dataclass
class DBModelRegistryParameters(BaseModelRegistryParameters):
    """Database model registry parameters."""

    database: Optional[BaseDatasourceParameters] = field(
        default=None, metadata={"help": _("Database configuration for model registry")}
    )


@dataclass
class ModelControllerParameters(BaseServerParameters):
    port: Optional[int] = field(
        default=8000, metadata={"help": _("Model Controller deploy port")}
    )
    registry: Optional[BaseModelRegistryParameters] = field(
        default=None,
        metadata={
            "help": _("Model registry configuration. If None, use embedded registry")
        },
    )

    heartbeat_interval_secs: Optional[int] = field(
        default=20,
        metadata={"help": _("The interval for checking heartbeats (seconds)")},
    )
    heartbeat_timeout_secs: Optional[int] = field(
        default=60,
        metadata={
            "help": _(
                "The timeout for checking heartbeats (seconds), it will be set "
                "unhealthy if the worker is not responding in this time"
            )
        },
    )


@dataclass
class ModelAPIServerParameters(BaseServerParameters):
    port: Optional[int] = field(
        default=8100, metadata={"help": _("Model API server deploy port")}
    )
    controller_addr: Optional[str] = field(
        default="http://127.0.0.1:8000",
        metadata={"help": _("The Model controller address to connect")},
    )

    api_keys: Optional[str] = field(
        default=None,
        metadata={"help": _("Optional list of comma separated API keys")},
    )
    embedding_batch_size: Optional[int] = field(
        default=None, metadata={"help": _("Embedding batch size")}
    )
    ignore_stop_exceeds_error: Optional[bool] = field(
        default=False, metadata={"help": _("Ignore exceeds stop words error")}
    )


@dataclass
class ModelWorkerParameters(BaseServerParameters):
    worker_type: Optional[str] = field(
        default=None,
        metadata={"valid_values": WorkerType.values(), "help": _("Worker type")},
    )
    worker_class: Optional[str] = field(
        default=None,
        metadata={
            "help": _("Model worker class, dbgpt.model.cluster.DefaultModelWorker")
        },
    )

    port: Optional[int] = field(
        default=8001, metadata={"help": _("Model worker deploy port")}
    )
    standalone: Optional[bool] = field(
        default=False,
        metadata={"help": _("Standalone mode. If True, embedded Run ModelController")},
    )
    register: Optional[bool] = field(
        default=True,
        metadata={"help": _("Register current worker to model controller")},
    )
    worker_register_host: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The ip address of current worker to register to ModelController. "
                "If None, the address is automatically determined"
            )
        },
    )
    controller_addr: Optional[str] = field(
        default=None, metadata={"help": _("The Model controller address to register")}
    )
    send_heartbeat: Optional[bool] = field(
        default=True, metadata={"help": _("Send heartbeat to model controller")}
    )
    heartbeat_interval: Optional[int] = field(
        default=20,
        metadata={"help": _("The interval for sending heartbeats (seconds)")},
    )


@dataclass
class ModelServiceConfig(BaseParameters):
    """Model service configuration."""

    __cfg_type__ = "service"

    worker: ModelWorkerParameters = field(
        default_factory=ModelWorkerParameters,
        metadata={"help": _("Model worker configuration")},
    )
    api: ModelAPIServerParameters = field(
        default_factory=ModelControllerParameters, metadata={"help": _("Model API")}
    )
    controller: ModelControllerParameters = field(
        default_factory=ModelControllerParameters,
        metadata={"help": _("Model controller")},
    )


@dataclass
class ModelsDeployParameters(BaseParameters):
    __cfg_type__ = "service"
    default_llm: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "Default LLM model name, used to specify which model to use when you "
                "have multiple LLMs"
            ),
        },
    )
    default_embedding: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "Default embedding model name, used to specify which model to use when "
                "you have multiple embedding models"
            ),
        },
    )
    default_reranker: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "Default reranker model name, used to specify which model to use when "
                "you have multiple reranker models"
            ),
        },
    )
    llms: List[LLMDeployModelParameters] = field(
        default_factory=list,
        metadata={
            "help": _(
                "LLM model deploy configuration. If you deploy in cluster mode, you "
                "just deploy one model."
            )
        },
    )
    embeddings: List[EmbeddingDeployModelParameters] = field(
        default_factory=list,
        metadata={
            "help": _(
                "Embedding model deploy configuration. If you deploy in cluster "
                "mode, you just deploy one model."
            )
        },
    )
    rerankers: List[RerankerDeployModelParameters] = field(
        default_factory=list,
        metadata={
            "help": _(
                "Reranker model deploy configuration. If you deploy in cluster "
                "mode, you just deploy one model."
            )
        },
    )

    def __post_init__(self):
        """Post init method."""
        llms, embeds, rerankers = [], [], []
        for llm in self.llms:
            llms.append(llm.name)
        for embedding in self.embeddings:
            embeds.append(embedding.name)
        for reranker in self.rerankers:
            rerankers.append(reranker.name)
        if not self.default_llm and llms:
            self.default_llm = llms[0]
        if not self.default_embedding and embeds:
            self.default_embedding = embeds[0]
        if not self.default_reranker and rerankers:
            self.default_reranker = rerankers[0]



import logging
from typing import Dict, List, Optional, Type, Union

from dbgpt.configs.model_config import get_device
from dbgpt.core import Embeddings, ModelMetadata, RerankEmbeddings
from dbgpt.core.interface.parameter import (
    BaseDeployModelParameters,
    EmbeddingDeployModelParameters,
    RerankerDeployModelParameters,
)
from dbgpt.model.adapter.base import EmbeddingModelAdapter, get_embedding_adapter
from dbgpt.model.cluster.worker_base import ModelWorker
from dbgpt.model.parameter import (
    WorkerType,
)
from dbgpt.util.model_utils import _clear_model_cache

logger = logging.getLogger(__name__)


class EmbeddingsModelWorker(ModelWorker):
    def __init__(self, rerank_model: bool = False) -> None:
        self._embeddings_impl: Union[Embeddings, RerankEmbeddings, None] = None
        self._model_params: Optional[
            Union[EmbeddingDeployModelParameters, RerankerDeployModelParameters]
        ] = None
        self._param_cls: Optional[
            Union[
                Type[EmbeddingDeployModelParameters],
                Type[RerankerDeployModelParameters],
            ]
        ] = None
        self._adapter: Optional[EmbeddingModelAdapter] = None

        self.model_name: str = ""
        self.model_path: str = ""
        self._rerank_model = rerank_model
        self._device = get_device()

    def load_worker(
        self,
        model_name: str,
        deploy_model_params: BaseDeployModelParameters,
        **kwargs,
    ) -> None:
        self.model_name = model_name
        if isinstance(
            deploy_model_params, EmbeddingDeployModelParameters
        ) or isinstance(deploy_model_params, RerankerDeployModelParameters):
            self._model_params = deploy_model_params
            self._param_cls = deploy_model_params.__class__
            self.model_path = deploy_model_params.real_provider_model_name
            self._adapter = get_embedding_adapter(
                deploy_model_params.provider,
                self._rerank_model,
                self.model_name,
                self.model_path,
            )
            if self._model_params.real_device:
                # Assign device from model params
                self._device = self._model_params.real_device
        else:
            raise ValueError(
                f"Invalid deploy_model_params type: {type(deploy_model_params)}"
            )

    def worker_type(self) -> WorkerType:
        return WorkerType.TEXT2VEC

    def model_param_class(self) -> Type[BaseDeployModelParameters]:
        return self._param_cls

    def start(
        self,
        command_args: List[str] = None,
    ) -> None:
        """Start model worker"""

        if self._rerank_model:
            logger.info(f"Load rerank embeddings model: {self.model_name}")
            self._embeddings_impl = self._adapter.load_from_params(self._model_params)
        else:
            logger.info(f"Load embeddings model: {self.model_name}")
            self._embeddings_impl = self._adapter.load_from_params(self._model_params)

    def __del__(self):
        self.stop()

    def stop(self) -> None:
        if not self._embeddings_impl:
            return
        del self._embeddings_impl
        self._embeddings_impl = None
        _clear_model_cache(self._device)

    def generate_stream(self, params: Dict):
        """Generate stream result, chat scene"""
        raise NotImplementedError("Not supported generate_stream for embeddings model")

    def generate(self, params: Dict):
        """Generate non stream result"""
        raise NotImplementedError("Not supported generate for embeddings model")

    def count_token(self, prompt: str) -> int:
        raise NotImplementedError("Not supported count_token for embeddings model")

    def get_model_metadata(self, params: Dict) -> ModelMetadata:
        raise NotImplementedError(
            "Not supported get_model_metadata for embeddings model"
        )

    def embeddings(self, params: Dict) -> List[List[float]]:
        model = params.get("model")
        logger.info(f"Receive embeddings request, model: {model}")
        textx: List[str] = params["input"]
        if isinstance(self._embeddings_impl, RerankEmbeddings):
            query = params["query"]
            scores: List[float] = self._embeddings_impl.predict(query, textx)
            return [scores]
        else:
            return self._embeddings_impl.embed_documents(textx)


class RerankerModelWorker(EmbeddingsModelWorker):
    def __init__(self) -> None:
        super().__init__(rerank_model=True)

    def worker_type(self) -> WorkerType:
        return WorkerType.RERANKER



import os
import shutil
import sys
import tempfile
from dataclasses import dataclass
from typing import Type

import pytest

from dbgpt.util.module_utils import ModelScanner, ScannerConfig


# Create test base classes and implementations
class TestBaseClass:
    """Base class for testing"""

    pass


class TestImplementation1(TestBaseClass):
    """First test implementation"""

    pass


class TestImplementation2(TestBaseClass):
    """Second test implementation"""

    pass


@dataclass
class TestParams:
    """Test parameters class for parameter scanning tests"""

    value: str


class TestParamsImpl1(TestParams):
    """First test parameters implementation"""

    pass


class TestParamsImpl2(TestParams):
    """Second test parameters implementation"""

    pass


# Test module content template
TEST_MODULE_CONTENT = '''
from typing import Optional
from {base_module} import TestBaseClass, TestParams

class {class_name}(TestBaseClass):
    """Test implementation in module"""
    pass

class {params_class_name}(TestParams):
    """Test parameters implementation in module"""
    value: Optional[str] = None
'''


class TestModelScanner:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        """Setup test environment and clean up after tests"""
        # Create temporary directory for test modules
        self.temp_dir = tempfile.mkdtemp()
        self.module_name = "test_modules"
        self.module_path = os.path.join(self.temp_dir, self.module_name)
        os.makedirs(self.module_path)

        # Create __init__.py
        with open(os.path.join(self.module_path, "__init__.py"), "w") as f:
            f.write("")

        # Add temporary directory to Python path
        sys.path.insert(0, self.temp_dir)

        yield

        # Cleanup
        sys.path.remove(self.temp_dir)
        shutil.rmtree(self.temp_dir)

        # Remove imported test modules from sys.modules
        for key in list(sys.modules.keys()):
            if key.startswith(self.module_name):
                del sys.modules[key]

    def create_test_module(
        self, module_name: str, class_name: str, params_class_name: str
    ):
        """Create a test module with test classes"""
        module_path = os.path.join(self.module_path, f"{module_name}.py")

        # Get the current module's name
        current_module = TestBaseClass.__module__

        with open(module_path, "w") as f:
            f.write(
                TEST_MODULE_CONTENT.format(
                    base_module=current_module,
                    class_name=class_name,
                    params_class_name=params_class_name,
                )
            )

    def test_basic_scanning(self):
        """Test basic class scanning functionality"""
        # Create test modules
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")
        self.create_test_module("module2", "TestImpl2", "TestParamsImpl2")

        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name, base_class=TestBaseClass, recursive=True
        )

        results = scanner.scan_and_register(config)

        # Check if classes were found
        assert len(results) == 2
        assert all(issubclass(cls, TestBaseClass) for cls in results.values())
        assert "test_modules.module1.testimpl1" in results
        assert "test_modules.module2.testimpl2" in results

    def test_recursive_scanning(self):
        """Test recursive directory scanning"""
        # Create nested directory structure
        nested_dir = os.path.join(self.module_path, "nested")
        os.makedirs(nested_dir)
        with open(os.path.join(nested_dir, "__init__.py"), "w") as f:
            f.write("")

        # Create test modules in different directories
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")
        self.create_test_module("nested/module2", "TestImpl2", "TestParamsImpl2")

        # Test with recursive=True
        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name, base_class=TestBaseClass, recursive=True
        )

        results = scanner.scan_and_register(config)
        assert len(results) == 2

        # Test with recursive=False
        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name, base_class=TestBaseClass, recursive=False
        )

        results = scanner.scan_and_register(config)
        assert len(results) == 1

    def test_custom_filter(self):
        """Test custom class filter functionality"""
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")
        self.create_test_module("module2", "TestImpl2", "TestParamsImpl2")

        def custom_filter(cls: Type) -> bool:
            return cls.__name__.endswith("1")

        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name,
            base_class=TestBaseClass,
            class_filter=custom_filter,
            recursive=True,
        )

        results = scanner.scan_and_register(config)
        assert len(results) == 1
        assert "test_modules.module1.testimpl1" in results

    def test_multiple_base_classes(self):
        """Test scanning for multiple different base classes"""
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")

        # Scan for TestBaseClass
        scanner1 = ModelScanner[TestBaseClass]()
        config1 = ScannerConfig(
            module_path=self.module_name, base_class=TestBaseClass, recursive=True
        )
        results1 = scanner1.scan_and_register(config1)

        # Scan for TestParams
        scanner2 = ModelScanner[TestParams]()
        config2 = ScannerConfig(
            module_path=self.module_name, base_class=TestParams, recursive=True
        )
        results2 = scanner2.scan_and_register(config2)

        assert len(results1) == 1
        assert len(results2) == 1
        assert "test_modules.module1.testimpl1" in results1
        assert "test_modules.module1.testparamsimpl1" in results2

    def test_error_handling(self):
        """Test error handling for invalid modules and paths"""
        scanner = ModelScanner[TestBaseClass]()

        # Test with non-existent module
        config = ScannerConfig(
            module_path="non_existent_module", base_class=TestBaseClass
        )
        results = scanner.scan_and_register(config)
        assert len(results) == 0

        # Test with invalid module content
        with open(os.path.join(self.module_path, "invalid_module.py"), "w") as f:
            f.write("This is invalid Python code!!!!")

        config = ScannerConfig(module_path=self.module_name, base_class=TestBaseClass)
        results = scanner.scan_and_register(config)
        assert len(results) == 0

    def test_specific_files(self):
        """Test scanning specific files"""
        # Create multiple test modules
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")
        self.create_test_module("module2", "TestImpl2", "TestParamsImpl2")
        self.create_test_module("module3", "TestImpl3", "TestParamsImpl3")

        # Test scanning specific single file
        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name,
            base_class=TestBaseClass,
            specific_files=["module1"],
        )
        results = scanner.scan_and_register(config)

        # Check results
        assert len(results) == 1
        assert "test_modules.module1.testimpl1" in results
        assert "testimpl2" not in results
        assert "testimpl3" not in results

        # Test scanning multiple specific files
        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name,
            base_class=TestBaseClass,
            specific_files=["module1", "module2"],
        )
        results = scanner.scan_and_register(config)

        # Check results
        assert len(results) == 2
        assert "test_modules.module1.testimpl1" in results
        assert "test_modules.module2.testimpl2" in results
        assert "testimpl3" not in results

    def test_specific_files_not_found(self):
        """Test scanning specific files that don't exist"""
        # Create one test module
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")

        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name,
            base_class=TestBaseClass,
            specific_files=["non_existent_module", "module1"],
        )
        results = scanner.scan_and_register(config)

        # Should still find the existing module
        assert len(results) == 1
        assert "test_modules.module1.testimpl1" in results

    def test_specific_files_in_nested_directory(self):
        """Test scanning specific files in nested directories"""
        # Create nested directory structure
        nested_dir = os.path.join(self.module_path, "nested")
        os.makedirs(nested_dir)
        with open(os.path.join(nested_dir, "__init__.py"), "w") as f:
            f.write("")

        # Create test modules in different directories
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")
        self.create_test_module("nested/module2", "TestImpl2", "TestParamsImpl2")

        # Test scanning specific file in nested directory
        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=f"{self.module_name}.nested",
            base_class=TestBaseClass,
            specific_files=["module2"],
        )
        results = scanner.scan_and_register(config)

        # Check results
        assert len(results) == 1
        assert "test_modules.nested.module2.testimpl2" in results

    def test_specific_files_with_filter(self):
        """Test scanning specific files with additional filter"""
        # Create test modules
        self.create_test_module("module1", "TestImpl1", "TestParamsImpl1")
        self.create_test_module("module2", "TestImpl2", "TestParamsImpl2")

        def custom_filter(cls: Type) -> bool:
            return cls.__name__.endswith("1")

        # Test scanning specific files with filter
        scanner = ModelScanner[TestBaseClass]()
        config = ScannerConfig(
            module_path=self.module_name,
            base_class=TestBaseClass,
            specific_files=["module1", "module2"],
            class_filter=custom_filter,
        )
        results = scanner.scan_and_register(config)

        # Should only find classes that match both criteria
        assert len(results) == 1
        assert "test_modules.module1.testimpl1" in results
        assert "testimpl2" not in results



"""Configuration base module."""

import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Optional, Union

logger = logging.getLogger(__name__)


class _MISSING_TYPE:
    pass


_MISSING = _MISSING_TYPE()


class ConfigCategory(str, Enum):
    """The configuration category."""

    AGENT = "agent"


class ProviderType(str, Enum):
    """The provider type."""

    ENV = "env"
    PROMPT_MANAGER = "prompt_manager"


class ConfigProvider(ABC):
    """The configuration provider."""

    name: ProviderType

    @abstractmethod
    def query(self, key: str, **kwargs) -> Any:
        """Query the configuration value by key."""


class EnvironmentConfigProvider(ConfigProvider):
    """Environment configuration provider.

    Obtain the configuration value from the environment variable.
    """

    name: ProviderType = ProviderType.ENV

    def query(self, key: str, **kwargs) -> Any:
        import os

        return os.environ.get(key, None)


class PromptManagerConfigProvider(ConfigProvider):
    """Prompt manager configuration provider.

    Obtain the configuration value from the prompt manager.

    It is valid only when DB-GPT web server is running for now.
    """

    name: ProviderType = ProviderType.PROMPT_MANAGER

    def query(self, key: str, **kwargs) -> Any:
        from dbgpt._private.config import Config

        try:
            from dbgpt_serve.prompt.serve import Serve
        except ImportError:
            logger.debug("Prompt manager is not available.")
            return None

        cfg = Config()
        sys_app = cfg.SYSTEM_APP
        if not sys_app:
            return None
        prompt_serve = Serve.get_instance(sys_app)
        if not prompt_serve or not prompt_serve.prompt_manager:
            return None
        prompt_manager = prompt_serve.prompt_manager
        value = prompt_manager.prefer_query(key, **kwargs)
        if not value:
            return None
        # Just return the first value
        return value[0].to_prompt_template().template


class ConfigInfo:
    def __init__(
        self,
        default: Any,
        key: Optional[str] = None,
        provider: Optional[Union[str, ConfigProvider]] = None,
        is_list: bool = False,
        separator: str = "[LIST_SEP]",
        description: Optional[str] = None,
    ):
        self.default = default
        self.key = key
        self.provider = provider
        self.is_list = is_list
        self.separator = separator
        self.description = description

    def query(self, **kwargs) -> Any:
        if self.key is None:
            return self.default
        value: Any = None
        if isinstance(self.provider, ConfigProvider):
            value = self.provider.query(self.key, **kwargs)
        elif self.provider == ProviderType.ENV:
            value = EnvironmentConfigProvider().query(self.key, **kwargs)
        elif self.provider == ProviderType.PROMPT_MANAGER:
            value = PromptManagerConfigProvider().query(self.key, **kwargs)
        if value is None:
            value = self.default
        if value and self.is_list and isinstance(value, str):
            value = value.split(self.separator)
        return value


def DynConfig(
    default: Any = _MISSING,
    *,
    category: str | ConfigCategory | None = None,
    key: str | None = None,
    provider: str | ProviderType | ConfigProvider | None = None,
    is_list: bool = False,
    separator: str = "[LIST_SEP]",
    description: str | None = None,
) -> Any:
    """Dynamic configuration.

    It allows to query the configuration value dynamically.
    It can obtain the configuration value from the specified provider.

    **Note**: Now just support obtaining string value or string list value.

    Args:
        default (Any): The default value.
        category (str | ConfigCategory | None): The configuration category.
        key (str | None): The configuration key.
        provider (str | ProviderType | ConfigProvider | None): The configuration
            provider.
        is_list (bool): Whether the value is a list.
        separator (str): The separator to split the list value.
        description (str | None): The configuration description.
    """
    if provider is None and category == ConfigCategory.AGENT:
        provider = ProviderType.PROMPT_MANAGER
    if default == _MISSING and key is None:
        raise ValueError("Default value or key is required.")
    if default != _MISSING and isinstance(default, list):
        is_list = True
    return ConfigInfo(
        default=default,
        key=key,
        provider=provider,
        is_list=is_list,
        separator=separator,
        description=description,
    )



"""Memory graph store."""

import logging

from dbgpt._private.pydantic import ConfigDict
from dbgpt.storage.graph_store.base import GraphStoreBase, GraphStoreConfig
from dbgpt.storage.graph_store.graph import MemoryGraph

logger = logging.getLogger(__name__)


class MemoryGraphStoreConfig(GraphStoreConfig):
    """Memory graph store config."""

    model_config = ConfigDict(arbitrary_types_allowed=True)


class MemoryGraphStore(GraphStoreBase):
    """Memory graph store."""

    def __init__(self, graph_store_config: MemoryGraphStoreConfig):
        """Initialize MemoryGraphStore with a memory graph."""
        self._graph_store_config = graph_store_config
        self._graph = MemoryGraph()

    def get_config(self):
        """Get the graph store config."""
        return self._graph_store_config



from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, Dict, List, Literal, Optional, Union

from dbgpt.core.schema.types import (
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
)
from dbgpt.util.i18n_utils import _

MEDIA_DATA_TYPE = Union[str, bytes]
MEDIA_DATA_FORMAT_TYPE = Literal[
    "text",
    "url",
    "base64",
    "binary",
]


@dataclass
class MediaObject:
    """Media object for the model output or model request."""

    data: MEDIA_DATA_TYPE = field(metadata={"help": _("The media data")})
    format: MEDIA_DATA_FORMAT_TYPE = field(
        default="text", metadata={"help": _("The format of the media")}
    )


class MediaContentType(str, Enum):
    """The media content type."""

    TEXT = "text"
    THINKING = "thinking"
    IMAGE = "image"
    AUDIO = "audio"
    VIDEO = "video"


@dataclass
class MediaContent:
    """Media content for the model output or model request.

    Examples:
        .. code-block:: python

        simple_text = MediaContent(
            type="text",
            object=MediaObject(
                data="Hello, world!",
                format="text",
            )
        )
        thinking_text = MediaContent(
            type="thinking",
            object=MediaObject(
                data="Thinking...",
                format="text",
            )
        )

        url_image1 = MediaContent(
            type="image",
            object=MediaObject(
                data="https://example.com/image.jpg",
                format="url",
            )
        )
        # Url with image type: 'image/jpeg'
        url_image2 = MediaContent(
            type="image",
            object=MediaObject(
                data="https://example.com/image.jpg",
                format="url@image/jpeg",
            )
        )

        # With image type: 'image/jpeg'
        base64_image1 = MediaContent(
            type="image",
            object=MediaObject(
                data="base64_string",
                format="base64@image/jpeg",
            )
        )
        # No image type
        base64_image2 = MediaContent(
            type="image",
            object=MediaObject(
                data="base64_string",
                format="base64",
            )
        )

        # Video
        url_video1 = MediaContent(
            type="video",
            object=MediaObject(
                data="https://example.com/video.mp4",
                format="url",
            )
        )
        url_video2 = MediaContent(
            type="video",
            object=MediaObject(
                data="https://example.com/video.mp4",
                format="url@video/mp4",
            )
        )
        binary_video = MediaContent(
            type="video",
            object=MediaObject(
                data=b"binary_data",
                format="binary@video/mp4",
            )
        )
        binary_audio = MediaContent(
            type="audio",
            object=MediaObject(
                data=b"binary_data",
                format="binary@audio/mpeg",
            )
        )
    """

    object: MediaObject = field(metadata={"help": _("The media object")})
    type: Literal["text", "thinking", "image", "audio", "video"] = field(
        default="text", metadata={"help": _("The type of the model media content")}
    )

    @classmethod
    def build_text(cls, text: str) -> "MediaContent":
        """Create a MediaContent object from text."""
        return cls(type="text", object=MediaObject(data=text, format="text"))

    @classmethod
    def build_thinking(cls, text: str) -> "MediaContent":
        """Create a MediaContent object from thinking."""
        return cls(type="thinking", object=MediaObject(data=text, format="text"))

    @classmethod
    def parse_content(
        cls,
        content: Union[
            "MediaContent", List["MediaContent"], Dict[str, Any], List[Dict[str, Any]]
        ],
    ) -> Union["MediaContent", List["MediaContent"]]:
        def _parse_dict(obj_dict: Union[MediaContent, Dict[str, Any]]) -> MediaContent:
            if isinstance(obj_dict, MediaContent):
                return obj_dict
            content_object = obj_dict.get("object")
            if not content_object:
                raise ValueError(f"Failed to parse {obj_dict}, no object found")
            if isinstance(content_object, dict):
                content_object = MediaObject(
                    data=content_object.get("data"),
                    format=content_object.get("format", "text"),
                )
            return cls(
                type=obj_dict.get("type", "text"),
                object=content_object,
            )

        if isinstance(content, list):
            return [_parse_dict(c) for c in content]
        else:
            return _parse_dict(content)

    def get_text(self) -> str:
        """Get the text."""
        if self.type == MediaContentType.TEXT:
            return self.object.data
        raise ValueError("The content type is not text")

    def get_thinking(self) -> str:
        """Get the thinking."""
        if self.type == MediaContentType.THINKING:
            return self.object.data
        raise ValueError("The content type is not thinking")

    @classmethod
    def last_text(cls, contents: List["MediaContent"]) -> str:
        """Get the last text from the contents."""
        if not contents:
            raise ValueError("The contents are empty")
        for content in reversed(contents):
            if content.type == MediaContentType.TEXT:
                return content.get_text()
        raise ValueError("No text content found")

    @classmethod
    def parse_chat_completion_message(
        cls,
        message: Union[str, ChatCompletionMessageParam],
        ignore_unknown_media: bool = False,
    ) -> Union["MediaContent", List["MediaContent"]]:
        """Parse the chat completion message."""
        if not message:
            raise ValueError("The message is empty")
        if isinstance(message, str):
            return cls.build_text(message)
        content = message.get("content")
        if not content:
            raise ValueError(f"Failed to parse {message}, no content found")
        if not isinstance(content, Iterable):
            raise ValueError(f"Failed to parse {message}, content is not iterable")
        result = []
        for item in content:
            if isinstance(item, str):
                result.append(cls.build_text(item))
            elif isinstance(item, dict) and "type" in item:
                type = item["type"]
                if type == "text" and "text" in item:
                    result.append(cls.build_text(item["text"]))
                elif type == "image_url" and "image_url" in item:
                    result.append(
                        cls(
                            type="image",
                            object=MediaObject(
                                data=item["image_url"]["url"],
                                format="url",
                            ),
                        )
                    )
                elif type == "input_audio" and "input_audio" in item:
                    result.append(
                        cls(
                            type="audio",
                            object=MediaObject(
                                data=item["input_audio"]["data"],
                                format="base64",
                            ),
                        )
                    )
                else:
                    if not ignore_unknown_media:
                        raise ValueError(
                            f"Unknown message type: {item} of system message"
                        )
            else:
                raise ValueError(f"Unknown message type: {item} of system message")
        return result

    @classmethod
    def to_chat_completion_message(
        cls,
        role,
        content: Union[str, "MediaContent", List["MediaContent"]],
        support_media_content: bool = True,
        type_mapping: Optional[Dict[str, str]] = None,
    ) -> ChatCompletionMessageParam:
        """Convert the media contents to chat completion message."""
        if not content:
            raise ValueError("The content are empty")
        if isinstance(content, str):
            return {"role": role, "content": content}
        if isinstance(content, MediaContent):
            content = [content]
        new_content = [
            cls._parse_single_media_content(c, type_mapping=type_mapping)
            for c in content
        ]
        if not support_media_content:
            text_content = [
                c["text"] for c in new_content if c["type"] == "text" and "text" in c
            ]
            if not text_content:
                raise ValueError("No text content found in the media contents")
            # Not support media content, just pass the string text as content
            new_content = text_content[0]
        return {
            "role": role,
            "content": new_content,
        }

    @classmethod
    def _parse_single_media_content(
        cls,
        content: "MediaContent",
        type_mapping: Optional[Dict[str, str]] = None,
    ) -> ChatCompletionContentPartParam:
        """Parse a single content."""
        if type_mapping is None:
            type_mapping = {}
        if content.type == MediaContentType.TEXT:
            real_type = type_mapping.get("text", "text")
            return {
                real_type: str(content.object.data),
                "type": real_type,
            }
        elif content.type == MediaContentType.IMAGE:
            if content.object.format.startswith("url"):
                # Compatibility for most image url formats
                real_type = type_mapping.get("image_url", "image_url")
                if real_type == "image_url":
                    return {
                        "image_url": {
                            "url": content.object.data,
                        },
                        "type": "image_url",
                    }
                else:
                    return {
                        real_type: content.object.data,
                        "type": real_type,
                    }
            else:
                raise ValueError(f"Unsupported image format: {content.object.format}")
        elif content.type == MediaContentType.AUDIO:
            if content.object.format.startswith("base64"):
                real_type = type_mapping.get("input_audio", "input_audio")
                return {
                    real_type: {
                        "data": content.object.data,
                    },
                    "type": real_type,
                }
            else:
                raise ValueError(f"Unsupported audio format: {content.object.format}")
        elif content.type == MediaContentType.VIDEO:
            if content.object.format.startswith("url"):
                real_type = type_mapping.get("video_url", "video_url")
                return {
                    real_type: {
                        "url": content.object.data,
                    },
                    "type": real_type,
                }
            else:
                raise ValueError(f"Unsupported video format: {content.object.format}")
        else:
            raise ValueError(f"Unsupported content type: {content.type}")

    @classmethod
    def get_view_markdown_text(
        cls,
        media_contents: List["MediaContent"],
        replace_url_func: Callable[[str], str],
    ) -> str:
        texts = []
        media = []
        for content in media_contents:
            if content.type == MediaContentType.TEXT:
                texts.append(content.get_text())
            elif content.type == MediaContentType.IMAGE:
                if content.object.format.startswith("url"):
                    media.append(f"![image]({replace_url_func(content.object.data)})")
                else:
                    raise ValueError(
                        f"Unsupported image format: {content.object.format}"
                    )
            elif content.type == MediaContentType.AUDIO:
                if content.object.format.startswith("base64"):
                    media.append(f"[audio](data:{content.object.data})")
                else:
                    raise ValueError(
                        f"Unsupported audio format: {content.object.format}"
                    )
            elif content.type == MediaContentType.VIDEO:
                if content.object.format.startswith("url"):
                    media.append(f"[video]({replace_url_func(content.object.data)})")
                else:
                    raise ValueError(
                        f"Unsupported video format: {content.object.format}"
                    )
        res = ""
        if texts:
            res += "\n".join(texts)
        if media:
            res += "\n" + "\n".join(media)
        return res

    @classmethod
    def apply_func(
        cls,
        media_contents: List["MediaContent"],
        func: Callable[[MediaObject], MediaObject],
    ) -> List["MediaContent"]:
        """Apply a function to the media contents."""
        if not media_contents:
            raise ValueError("The media contents are empty")
        return [
            cls(type=content.type, object=func(content.object))
            for content in media_contents
        ]

    @classmethod
    def replace_url(
        cls,
        media_contents: List["MediaContent"],
        replace_func: Callable[[str], str],
    ) -> List["MediaContent"]:
        """Replace the url in the media contents."""
        if not media_contents:
            raise ValueError("The media contents are empty")

        def _func(obj: MediaObject) -> MediaObject:
            if obj.format.startswith("url"):
                data = replace_func(obj.data)
            else:
                data = obj.data
            return MediaObject(
                data=data,
                format=obj.format,
            )

        return cls.apply_func(media_contents, _func)


class MediaProcessor(ABC):
    """Media processor for the model output or model request."""

    @abstractmethod
    def process(self, content: MediaContent) -> MediaContent:
        """Process the media content."""

    def batch_process(self, contents: List[MediaContent]) -> List[MediaContent]:
        """Batch process the media content."""
        return [self.process(content) for content in contents]



import json
import re
from dataclasses import dataclass
from typing import Any, List, Optional


@dataclass
class ReActStep:
    """
    Dataclass representing a single step in the ReAct pattern.
    """

    thought: Optional[str] = None
    phase: Optional[str] = None
    action: Optional[str] = None
    action_input: Optional[Any] = None
    observation: Optional[Any] = None
    is_terminal: bool = False


class ReActOutputParser:
    """
    Parser for ReAct format model outputs with configurable prefixes.
    This parser extracts structured information from language model outputs
    that follow the ReAct pattern: Thought -> Phase -> Action -> Action Input
    -> Observation.
    """

    def __init__(
        self,
        thought_prefix: str = "Thought:",
        phase_prefix: str = "Phase:",
        action_prefix: str = "Action:",
        action_input_prefix: str = "Action Input:",
        observation_prefix: str = "Observation:",
        terminate_action: str = "terminate",
    ):
        """
        Initialize the ReAct output parser with configurable prefixes.

        Args:
            thought_prefix: Prefix string that indicates the start of a thought.
            phase_prefix: Prefix string that indicates the start of a phase.
            action_prefix: Prefix string that indicates the start of an action.
            action_input_prefix: Prefix string that indicates the start of action input.
            observation_prefix: Prefix string that indicates the start of an
                observation.
            terminate_action: String that indicates termination action.
        """
        self.thought_prefix = thought_prefix
        self.phase_prefix = phase_prefix
        self.action_prefix = action_prefix
        self.action_input_prefix = action_input_prefix
        self.observation_prefix = observation_prefix
        self.terminate_action = terminate_action

        # Escape special regex characters in prefixes
        self.thought_prefix_escaped = re.escape(thought_prefix)
        self.phase_prefix_escaped = re.escape(phase_prefix)
        self.action_prefix_escaped = re.escape(action_prefix)
        self.action_input_prefix_escaped = re.escape(action_input_prefix)
        self.observation_prefix_escaped = re.escape(observation_prefix)

    def parse(self, text: str) -> List[ReActStep]:
        """
        Parse the ReAct format output text into structured steps.

        Args:
            text: The text to parse, containing ReAct formatted content.

        Returns:
            List of ReActStep dataclasses, each containing thought, action,
                action_input, and observation.
        """
        # Split the text into steps based on thought prefix
        steps = []

        # Remove any leading/trailing whitespace
        text = text.strip()

        # Find all instances of the thought prefix
        thought_matches = list(re.finditer(rf"{self.thought_prefix_escaped}\s*", text))

        if not thought_matches:
            return []

        # Process each thought section
        for i, match in enumerate(thought_matches):
            start_pos = match.start()

            # Determine end position (either next thought or end of text)
            if i < len(thought_matches) - 1:
                end_pos = thought_matches[i + 1].start()
            else:
                end_pos = len(text)

            # Extract the current step's text
            step_text = text[start_pos:end_pos].strip()

            # Parse the step
            step_data = self._parse_step(step_text)
            if step_data:
                steps.append(step_data)

        return steps

    def _parse_step(self, step_text: str) -> Optional[ReActStep]:
        """
        Parse a single step of the ReAct format.

        Args:
            step_text: Text containing a single thought-action-input-observation
                sequence.

        Returns:
            ReActStep dataclass with thought, action, action_input, and observation,
                or None if parsing fails.
        """
        # Initialize the result
        thought = None
        action = None
        action_input = None
        observation = None
        is_terminal = False

        # Extract thought
        thought_match = re.search(
            rf"{self.thought_prefix_escaped}\s*(.*?)(?={self.phase_prefix_escaped}|{self.action_prefix_escaped}|{self.observation_prefix_escaped}|$)",
            step_text,
            re.DOTALL,
        )
        if thought_match:
            thought = thought_match.group(1).strip()

        # Extract phase (optional, between thought and action)
        phase = None
        phase_match = re.search(
            rf"{self.phase_prefix_escaped}\s*(.*?)(?={self.action_prefix_escaped}|{self.action_input_prefix_escaped}|{self.observation_prefix_escaped}|$)",
            step_text,
            re.DOTALL,
        )
        if phase_match:
            phase = phase_match.group(1).strip() or None

        # Extract action
        action_match = re.search(
            rf"{self.action_prefix_escaped}\s*(.*?)(?={self.action_input_prefix_escaped}|{self.observation_prefix_escaped}|$)",
            step_text,
            re.DOTALL,
        )
        if action_match:
            action = action_match.group(1).strip()

            # Check if this is a terminate action
            is_terminal = action.lower() == self.terminate_action.lower()

        # Extract action input
        action_input_match = re.search(
            rf"{self.action_input_prefix_escaped}\s*(.*?)(?={self.observation_prefix_escaped}|{self.thought_prefix_escaped}|$)",
            step_text,
            re.DOTALL,
        )
        if action_input_match:
            action_input_text = action_input_match.group(1).strip()

            # Try to parse action input as JSON if it looks like JSON
            if (
                action_input_text.startswith("{") and action_input_text.endswith("}")
            ) or (
                action_input_text.startswith("[") and action_input_text.endswith("]")
            ):
                try:
                    action_input = json.loads(action_input_text)
                except json.JSONDecodeError:
                    action_input = action_input_text
            else:
                action_input = action_input_text

        # Extract observation
        observation_match = re.search(
            rf"{self.observation_prefix_escaped}\s*(.*?)(?={self.thought_prefix_escaped}|$)",
            step_text,
            re.DOTALL,
        )
        if observation_match:
            observation_text = observation_match.group(1).strip()

            # Try to parse observation as JSON if it looks like JSON
            if (
                observation_text.startswith("{") and observation_text.endswith("}")
            ) or (observation_text.startswith("[") and observation_text.endswith("]")):
                try:
                    observation = json.loads(observation_text)
                except json.JSONDecodeError:
                    observation = observation_text
            else:
                observation = observation_text

        # Only return if we have at least thought or action
        if thought or action:
            return ReActStep(
                thought=thought,
                phase=phase,
                action=action,
                action_input=action_input,
                observation=observation,
                is_terminal=is_terminal,
            )
        return None

    def get_final_output(self, steps: List[ReActStep]) -> Optional[str]:
        """
        Get the final output from a terminate action if it exists.

        Args:
            steps: List of parsed steps.

        Returns:
            The final output string or None if no terminate action is found.
        """
        for step in reversed(steps):  # Look from the end
            if step.is_terminal and step.action == self.terminate_action:
                if (
                    isinstance(step.action_input, dict)
                    and "output" in step.action_input
                ):
                    return step.action_input["output"]
        return None



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import math
from typing import Optional, Tuple

import torch
import torch.nn.functional as F
import transformers
from torch import nn


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2].clone()
    x2 = x[..., x.shape[-1] // 2 :].clone()
    return torch.cat((-x2, x1), dim=-1)


def apply_rotary_pos_emb(q, k, cos, sin, position_ids):
    gather_indices = position_ids[:, None, :, None]  # [bs, 1, seq_len, 1]
    gather_indices = gather_indices.repeat(1, cos.shape[1], 1, cos.shape[3])
    cos = torch.gather(cos.repeat(gather_indices.shape[0], 1, 1, 1), 2, gather_indices)
    sin = torch.gather(sin.repeat(gather_indices.shape[0], 1, 1, 1), 2, gather_indices)
    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed


def forward(
    self,
    hidden_states: torch.Tensor,
    attention_mask: Optional[torch.Tensor] = None,
    position_ids: Optional[torch.LongTensor] = None,
    past_key_value: Optional[Tuple[torch.Tensor]] = None,
    output_attentions: bool = False,
    use_cache: bool = False,
    padding_mask: Optional[torch.LongTensor] = None,
) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
    bsz, q_len, _ = hidden_states.size()

    query_states = (
        self.q_proj(hidden_states)
        .view(bsz, q_len, self.num_heads, self.head_dim)
        .transpose(1, 2)
    )
    key_states = (
        self.k_proj(hidden_states)
        .view(bsz, q_len, self.num_heads, self.head_dim)
        .transpose(1, 2)
    )
    value_states = (
        self.v_proj(hidden_states)
        .view(bsz, q_len, self.num_heads, self.head_dim)
        .transpose(1, 2)
    )

    kv_seq_len = key_states.shape[-2]
    if past_key_value is not None:
        kv_seq_len += past_key_value[0].shape[-2]
    cos, sin = self.rotary_emb(value_states, seq_len=kv_seq_len)
    query_states, key_states = apply_rotary_pos_emb(
        query_states, key_states, cos, sin, position_ids
    )
    # [bsz, nh, t, hd]

    if past_key_value is not None:
        # reuse k, v, self_attention
        key_states = torch.cat([past_key_value[0], key_states], dim=2)
        value_states = torch.cat([past_key_value[1], value_states], dim=2)

    past_key_value = (key_states, value_states) if use_cache else None

    attn_weights = torch.matmul(query_states, key_states.transpose(2, 3)) / math.sqrt(
        self.head_dim
    )

    if attn_weights.size() != (bsz, self.num_heads, q_len, kv_seq_len):
        raise ValueError(
            f"Attention weights should be of size "
            f"{(bsz * self.num_heads, q_len, kv_seq_len)}, but is"
            f" {attn_weights.size()}"
        )

    if attention_mask is not None:
        if attention_mask.size() != (bsz, 1, q_len, kv_seq_len):
            raise ValueError(
                f"Attention mask should be of size "
                f"{(bsz, 1, q_len, kv_seq_len)}, but is {attention_mask.size()}"
            )
        attn_weights = attn_weights + attention_mask
        attn_weights = torch.max(
            attn_weights, torch.tensor(torch.finfo(attn_weights.dtype).min)
        )

    # upcast attention to fp32
    attn_weights = nn.functional.softmax(attn_weights, dim=-1, dtype=torch.float32).to(
        query_states.dtype
    )
    attn_output = torch.matmul(attn_weights, value_states)

    if attn_output.size() != (bsz, self.num_heads, q_len, self.head_dim):
        raise ValueError(
            "`attn_output` should be of size "
            f"{(bsz, self.num_heads, q_len, self.head_dim)}, but is"
            f" {attn_output.size()}"
        )

    attn_output = attn_output.transpose(1, 2)
    attn_output = attn_output.reshape(bsz, q_len, self.hidden_size)

    attn_output = self.o_proj(attn_output)

    if not output_attentions:
        attn_weights = None

    return attn_output, attn_weights, past_key_value


def replace_llama_attn_with_non_inplace_operations():
    """Avoid bugs in mps backend by not using in-place operations."""
    transformers.models.llama.modeling_llama.LlamaAttention.forward = forward


class ParQwen3MoeSparseMoeBlock(nn.Module):
    """
    Adapted from https://huggingface.co/Qwen/Qwen3-30B-A3B/discussions/13
    """

    def __init__(self, base_moe):
        super().__init__()
        self.base_moe = base_moe
        self.num_experts = base_moe.num_experts
        self.top_k = base_moe.top_k
        self.norm_topk_prob = base_moe.norm_topk_prob

        # gating
        self.gate = base_moe.gate
        self.experts = base_moe.experts

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        batch_size, sequence_length, hidden_dim = hidden_states.shape
        hidden_states = hidden_states.view(-1, hidden_dim)
        # router_logits: (batch * sequence_length, n_experts)
        router_logits = self.gate(hidden_states)

        routing_weights = F.softmax(router_logits, dim=1, dtype=torch.float)
        routing_weights, selected_experts = torch.topk(
            routing_weights, self.top_k, dim=-1
        )
        if self.norm_topk_prob:  # only diff with mixtral sparse moe block!
            routing_weights /= routing_weights.sum(dim=-1, keepdim=True)
        # we cast back to the input dtype
        routing_weights = routing_weights.to(hidden_states.dtype)

        final_hidden_states = torch.zeros(
            (batch_size * sequence_length, hidden_dim),
            dtype=hidden_states.dtype,
            device=hidden_states.device,
        )

        # One hot encode the selected experts to create an expert mask
        # this will be used to easily index which expert is going to be sollicitated
        expert_mask = torch.nn.functional.one_hot(
            selected_experts, num_classes=self.num_experts
        ).permute(2, 1, 0)

        activated_experts = torch.unique(selected_experts)
        cuda_streams = [torch.cuda.Stream() for _ in activated_experts]
        # Loop over all available experts in the model and perform the computation on
        # each expert
        for expert_idx, cuda_stream in zip(activated_experts, cuda_streams):
            with torch.cuda.stream(cuda_stream):
                expert_layer = self.experts[expert_idx]
                idx, top_x = torch.where(expert_mask[expert_idx])

                # Index the correct hidden states and compute the expert hidden state
                # for the current expert. We need to make sure to multiply the output
                # hidden states by `routing_weights` on the corresponding tokens
                # (top-1 and top-2)
                current_state = hidden_states[None, top_x].reshape(-1, hidden_dim)
                current_hidden_states = (
                    expert_layer(current_state) * routing_weights[top_x, idx, None]
                )

                # However `index_add_` only support torch tensors for indexing so
                # we'll use the `top_x` tensor here.
                final_hidden_states.index_add_(
                    0, top_x, current_hidden_states.to(hidden_states.dtype)
                )
        torch.cuda.synchronize()
        final_hidden_states = final_hidden_states.reshape(
            batch_size, sequence_length, hidden_dim
        )
        return final_hidden_states, router_logits


def apply_qwen3_moe_monkey_patch(model):
    if not torch.cuda.is_available():
        # Only apply monkey patch if CUDA is available
        return model

    for layer in model.model.layers:
        if type(layer.mlp).__name__ == "Qwen3MoeSparseMoeBlock":
            layer.mlp = ParQwen3MoeSparseMoeBlock(layer.mlp)
    return model


def recovery_moe_monkey_patch(model):
    for layer in model.model.layers:
        if type(layer.mlp).__name__ == "ParQwen3MoeSparseMoeBlock" and hasattr(
            layer.mlp, "base_moe"
        ):
            layer.mlp = layer.mlp.base_moe
    return model



import logging
from concurrent.futures import Executor
from dataclasses import dataclass, field
from typing import Iterator, Optional, Type, Union

from dbgpt.core import MessageConverter, ModelMetadata, ModelOutput, ModelRequest
from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.core.interface.parameter import LLMDeployModelParameters
from dbgpt.model.proxy.base import (
    AsyncGenerateStreamFunction,
    GenerateStreamFunction,
    ProxyLLMClient,
    register_proxy_model_adapter,
)
from dbgpt.model.proxy.llms.proxy_model import ProxyModel, parse_model_request
from dbgpt.util.i18n_utils import _

from ...utils.parse_utils import (
    parse_chat_message,
)

logger = logging.getLogger(__name__)


@auto_register_resource(
    label=_("Ollama Proxy LLM"),
    category=ResourceCategory.LLM_CLIENT,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("Ollama proxy LLM configuration."),
    documentation_url="https://ollama.com/library",
    show_in_ui=False,
)
@dataclass
class OllamaDeployModelParameters(LLMDeployModelParameters):
    """Deploy model parameters for Ollama."""

    provider: str = "proxy/ollama"

    api_base: Optional[str] = field(
        default="${env:OLLAMA_API_BASE:-http://localhost:11434}",
        metadata={
            "help": _("The base url of the Ollama API."),
        },
    )


def ollama_generate_stream(
    model: ProxyModel, tokenizer, params, device, context_len=4096
):
    client: OllamaLLMClient = model.proxy_llm_client
    request = parse_model_request(params, client.default_model, stream=True)
    for r in client.sync_generate_stream(request):
        yield r


class OllamaLLMClient(ProxyLLMClient):
    def __init__(
        self,
        model: Optional[str] = None,
        api_base: Optional[str] = None,
        model_alias: Optional[str] = "llama2",
        context_length: Optional[int] = 4096,
        executor: Optional[Executor] = None,
    ):
        if not model:
            model = "llama2"
        if not api_base:
            api_base = "http://localhost:11434"
        self._model = model
        self._api_base = self._resolve_env_vars(api_base)

        super().__init__(
            model_names=[model, model_alias],
            context_length=context_length,
            executor=executor,
        )

    @classmethod
    def new_client(
        cls,
        model_params: OllamaDeployModelParameters,
        default_executor: Optional[Executor] = None,
    ) -> "OllamaLLMClient":
        return cls(
            model=model_params.real_provider_model_name,
            api_base=model_params.api_base,
            model_alias=model_params.real_provider_model_name,
            context_length=model_params.context_length,
            executor=default_executor,
        )

    @classmethod
    def param_class(cls) -> Type[OllamaDeployModelParameters]:
        return OllamaDeployModelParameters

    @classmethod
    def generate_stream_function(
        cls,
    ) -> Optional[Union[GenerateStreamFunction, AsyncGenerateStreamFunction]]:
        return ollama_generate_stream

    @property
    def default_model(self) -> str:
        return self._model

    def sync_generate_stream(
        self,
        request: ModelRequest,
        message_converter: Optional[MessageConverter] = None,
    ) -> Iterator[ModelOutput]:
        try:
            import ollama
            from ollama import Client
        except ImportError as e:
            raise ValueError(
                "Could not import python package: ollama "
                "Please install ollama by command `pip install ollama"
            ) from e
        request = self.local_covert_message(request, message_converter)
        messages = request.to_common_messages()

        model = request.model or self._model
        is_reasoning_model = getattr(request.context, "is_reasoning_model", False)
        client = Client(self._api_base)
        try:
            stream = client.chat(
                model=model,
                messages=messages,
                stream=True,
            )
            content = ""
            for chunk in stream:
                content = content + chunk["message"]["content"]
                msg = parse_chat_message(content, extract_reasoning=is_reasoning_model)
                yield ModelOutput.build(
                    text=msg.content, thinking=msg.reasoning_content, error_code=0
                )
        except ollama.ResponseError as e:
            yield ModelOutput.build(
                text=f"**Ollama Response Error, Please CheckErrorInfo.**: {e}",
                error_code=-1,
            )


register_proxy_model_adapter(
    OllamaLLMClient,
    supported_models=[
        ModelMetadata(
            model="deepseek-v3",
            context_length=64 * 1024,
            max_output_length=8 * 1024,
            description="DeepSeek-V3 by DeepSeek",
            link="https://ollama.com/library/deepseek-v3",
            function_calling=True,
        ),
        ModelMetadata(
            model="deepseek-r1:671b",
            context_length=64 * 1024,
            max_output_length=8 * 1024,
            description="DeepSeek-R1 by DeepSeek",
            link="https://ollama.com/library/deepseek-r1",
            function_calling=True,
        ),
        # More models see: https://ollama.com/search
    ],
)



"""Hooks for setting environment variables."""

import os
from typing import Any, Dict, List, Optional


class EnvVarSetHook:
    """A hook that sets environment variables based on initialization parameters"""

    def __init__(self, env_vars: Optional[List[Dict[str, str]]] = None):
        """
        Args:
            env_vars: Dictionary of environment variables to set.
                     Key is the environment variable name, value is its value.
        """
        env_kv = {}
        for env_var in env_vars or []:
            if not isinstance(env_var, dict):
                raise ValueError(
                    f"Expected env_vars to be a list of dictionaries, got {env_var}"
                )
            if not env_var:
                raise ValueError("Expected env_var to be a non-empty dictionary")
            env_key = env_var.get("key")
            env_value = env_var.get("value")
            env_kv[env_key] = env_value
        self.env_vars = env_kv
        self._original_env = {}

    def __call__(self, config: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """Set environment variables and return the config unchanged."""
        # Save original environment variables that we're going to override
        self._original_env = {
            key: os.environ.get(key) for key in self.env_vars if key in os.environ
        }

        # Set new environment variables
        for key, value in self.env_vars.items():
            os.environ[key] = str(value)

        return config



from typing import List

import pytest

from dbgpt.core.interface.message import AIMessage, BaseMessage, HumanMessage
from dbgpt.core.operators import BufferedConversationMapperOperator


@pytest.fixture
def messages() -> List[BaseMessage]:
    return [
        HumanMessage(content="Hi", round_index=1),
        AIMessage(content="Hello!", round_index=1),
        HumanMessage(content="How are you?", round_index=2),
        AIMessage(content="I'm good, thanks!", round_index=2),
        HumanMessage(content="What's new today?", round_index=3),
        AIMessage(content="Lots of things!", round_index=3),
    ]


@pytest.mark.asyncio
async def test_buffered_conversation_keep_start_rounds(messages: List[BaseMessage]):
    # Test keep_start_rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=2,
        keep_end_rounds=None,
    )
    assert await operator.map_messages(messages) == [
        HumanMessage(content="Hi", round_index=1),
        AIMessage(content="Hello!", round_index=1),
        HumanMessage(content="How are you?", round_index=2),
        AIMessage(content="I'm good, thanks!", round_index=2),
    ]
    # Test keep start 0 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=0,
        keep_end_rounds=None,
    )
    assert await operator.map_messages(messages) == []

    # Test keep start 100 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=100,
        keep_end_rounds=None,
    )
    assert await operator.map_messages(messages) == messages

    # Test keep start -1 rounds
    with pytest.raises(ValueError):
        operator = BufferedConversationMapperOperator(
            keep_start_rounds=-1,
            keep_end_rounds=None,
        )
        await operator.map_messages(messages)


@pytest.mark.asyncio
async def test_buffered_conversation_keep_end_rounds(messages: List[BaseMessage]):
    # Test keep_end_rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=None,
        keep_end_rounds=2,
    )
    assert await operator.map_messages(messages) == [
        HumanMessage(content="How are you?", round_index=2),
        AIMessage(content="I'm good, thanks!", round_index=2),
        HumanMessage(content="What's new today?", round_index=3),
        AIMessage(content="Lots of things!", round_index=3),
    ]
    # Test keep end 0 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=0,
        keep_end_rounds=0,
    )
    assert await operator.map_messages(messages) == []

    # Test keep end 100 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=None,
        keep_end_rounds=100,
    )
    assert await operator.map_messages(messages) == messages

    # Test keep end -1 rounds
    with pytest.raises(ValueError):
        operator = BufferedConversationMapperOperator(
            keep_start_rounds=None,
            keep_end_rounds=-1,
        )
        await operator.map_messages(messages)


@pytest.mark.asyncio
async def test_buffered_conversation_keep_start_end_rounds(messages: List[BaseMessage]):
    # Test keep_start_rounds and keep_end_rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=1,
        keep_end_rounds=1,
    )
    assert await operator.map_messages(messages) == [
        HumanMessage(content="Hi", round_index=1),
        AIMessage(content="Hello!", round_index=1),
        HumanMessage(content="What's new today?", round_index=3),
        AIMessage(content="Lots of things!", round_index=3),
    ]
    # Test keep start 0 rounds and keep end 0 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=0,
        keep_end_rounds=0,
    )
    assert await operator.map_messages(messages) == []

    # Test keep start 0 rounds and keep end 1 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=0,
        keep_end_rounds=1,
    )
    assert await operator.map_messages(messages) == [
        HumanMessage(content="What's new today?", round_index=3),
        AIMessage(content="Lots of things!", round_index=3),
    ]

    # Test keep start 2 rounds and keep end 0 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=2,
        keep_end_rounds=0,
    )
    assert await operator.map_messages(messages) == [
        HumanMessage(content="Hi", round_index=1),
        AIMessage(content="Hello!", round_index=1),
        HumanMessage(content="How are you?", round_index=2),
        AIMessage(content="I'm good, thanks!", round_index=2),
    ]

    # Test keep start 100 rounds and keep end 100 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=100,
        keep_end_rounds=100,
    )
    assert await operator.map_messages(messages) == messages

    # Test keep start 2 round and keep end 2 rounds
    operator = BufferedConversationMapperOperator(
        keep_start_rounds=2,
        keep_end_rounds=2,
    )
    assert await operator.map_messages(messages) == messages

    # Test keep start -1 rounds and keep end -1 rounds
    with pytest.raises(ValueError):
        operator = BufferedConversationMapperOperator(
            keep_start_rounds=-1,
            keep_end_rounds=-1,
        )
        await operator.map_messages(messages)



"""Vis Anomaly Detection."""

from typing import Any, Dict, Optional

from ..base import Vis


class VisAnomalyDetection(Vis):
    """VisAnomalyDetection."""

    @classmethod
    def vis_tag(cls):
        """Vis tag name.

        Returns:
            str: The tag name associated with the visualization.
        """
        return "anomaly-detection"

    async def display(self, content: Dict[str, Any]) -> Optional[str]:
        """Display the visualization content."""
        if not content:
            return None

        is_anomaly = content.get("is_anomaly", False)
        metric_name = content.get("metric_name", "Unknown Metric")
        fluctuation_rate = content.get("fluctuation_rate", 0)
        threshold = content.get("threshold", 0)

        if is_anomaly:
            anomaly_type = content.get("anomaly_type", "unknown")
            direction = "上升" if anomaly_type == "increase" else "下降"
            result = f"[异常检测] {metric_name} 检测到{direction}异常\n"
            result += f"  波动率: {fluctuation_rate * 100:.2f}%\n"
            result += f"  阈值: {threshold * 100:.2f}%\n"
        else:
            result = f"[异常检测] {metric_name} 未检测到异常\n"
            result += f"  波动率: {fluctuation_rate * 100:.2f}%\n"
            result += f"  阈值: {threshold * 100:.2f}%\n"

        return result



"""Profile management subcommands: list, show, create, switch, delete."""

import click


@click.group()
def profile():
    """Manage DB-GPT configuration profiles."""
    pass


@profile.command(name="list")
def profile_list():
    """List all configured profiles."""
    from dbgpt.cli._config import configs_dir, read_active_profile

    configs = configs_dir()
    active = read_active_profile()
    toml_files = sorted(configs.glob("*.toml"))

    if not toml_files:
        click.echo("No profiles configured. Run: dbgpt setup")
        return

    for f in toml_files:
        name = f.stem
        marker = "* " if name == active else "  "
        click.echo(f"{marker}{name}")


@profile.command()
@click.argument("name")
def show(name):
    """Show the TOML configuration for a profile."""
    from dbgpt.cli._config import profile_config_path

    path = profile_config_path(name)
    if not path.exists():
        raise click.ClickException(
            f"Profile '{name}' not found. Run: dbgpt profile create {name}"
        )
    click.echo(path.read_text(encoding="utf-8"))


@profile.command()
@click.argument("name")
def create(name):
    """Create or reconfigure a profile (runs the setup wizard)."""
    from dbgpt.cli._wizard import run_setup_wizard

    run_setup_wizard(pre_selected_profile=name)


@profile.command()
@click.argument("name")
def switch(name):
    """Set a profile as the active default."""
    from dbgpt.cli._config import profile_config_path, write_active_profile

    path = profile_config_path(name)
    if not path.exists():
        click.echo(
            f"Error: Profile '{name}' not found. Run: dbgpt profile create {name}",
            err=False,
        )
        raise SystemExit(1)
    write_active_profile(name)
    click.echo(f"Switched active profile to: {name}")


@profile.command()
@click.argument("name")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
def delete(name, yes):
    """Delete a profile configuration file."""
    from dbgpt.cli._config import (
        profile_config_path,
        read_active_profile,
        write_active_profile,
    )

    path = profile_config_path(name)
    if not path.exists():
        raise click.ClickException(f"Profile '{name}' not found.")

    if not yes:
        click.confirm(f"Delete profile '{name}'?", abort=True)

    path.unlink()

    # Clear active pointer if this was the active profile
    if read_active_profile() == name:
        write_active_profile("")  # clear by writing empty string

    click.echo(f"Deleted profile: {name}")



import asyncio
import json
import os
import tempfile
import time
from datetime import datetime, timedelta
from unittest.mock import patch

import pytest

from dbgpt.util.tracer import (
    FileSpanStorage,
    Span,
    SpanStorage,
    SpanStorageContainer,
    SpanType,
)


@pytest.fixture
def storage(request):
    if not request or not hasattr(request, "param"):
        file_does_not_exist = False
    else:
        file_does_not_exist = request.param.get("file_does_not_exist", False)

    if file_does_not_exist:
        with tempfile.TemporaryDirectory() as tmp_dir:
            filename = os.path.join(tmp_dir, "non_existent_file.jsonl")
            storage_instance = FileSpanStorage(filename)
            yield storage_instance
    else:
        with tempfile.NamedTemporaryFile(delete=True) as tmp_file:
            filename = tmp_file.name
            storage_instance = FileSpanStorage(filename)
            yield storage_instance


@pytest.fixture
def storage_container(request):
    if not request or not hasattr(request, "param"):
        batch_size = 10
        flush_interval = 10
    else:
        batch_size = request.param.get("batch_size", 10)
        flush_interval = request.param.get("flush_interval", 10)
    storage_container = SpanStorageContainer(
        batch_size=batch_size, flush_interval=flush_interval
    )
    yield storage_container


def read_spans_from_file(filename):
    with open(filename, "r") as f:
        return [json.loads(line) for line in f.readlines()]


def test_write_span(storage: SpanStorage):
    span = Span("1", "a", SpanType.BASE, "b", "op1")
    storage.append_span(span)
    time.sleep(0.1)

    spans_in_file = read_spans_from_file(storage.filename)
    assert len(spans_in_file) == 1
    assert spans_in_file[0]["trace_id"] == "1"


def test_incremental_write(storage: SpanStorage):
    span1 = Span("1", "a", SpanType.BASE, "b", "op1")
    span2 = Span("2", "c", SpanType.BASE, "d", "op2")

    storage.append_span(span1)
    storage.append_span(span2)
    time.sleep(0.1)

    spans_in_file = read_spans_from_file(storage.filename)
    assert len(spans_in_file) == 2


def test_sync_and_async_append(storage: SpanStorage):
    span = Span("1", "a", SpanType.BASE, "b", "op1")

    storage.append_span(span)

    async def async_append():
        storage.append_span(span)

    asyncio.run(async_append())

    time.sleep(0.1)
    spans_in_file = read_spans_from_file(storage.filename)
    assert len(spans_in_file) == 2


@pytest.mark.parametrize("storage", [{"file_does_not_exist": True}], indirect=True)
def test_non_existent_file(storage: SpanStorage):
    span = Span("1", "a", SpanType.BASE, "b", "op1")
    span2 = Span("2", "c", SpanType.BASE, "d", "op2")
    storage.append_span(span)
    time.sleep(0.1)

    spans_in_file = read_spans_from_file(storage.filename)
    assert len(spans_in_file) == 1

    storage.append_span(span2)
    time.sleep(0.1)
    spans_in_file = read_spans_from_file(storage.filename)
    assert len(spans_in_file) == 2
    assert spans_in_file[0]["trace_id"] == "1"
    assert spans_in_file[1]["trace_id"] == "2"


@pytest.mark.parametrize("storage", [{"file_does_not_exist": True}], indirect=True)
def test_log_rollover(storage: SpanStorage):
    # mock start date
    mock_start_date = datetime(2023, 10, 18, 23, 59)

    with patch("datetime.datetime") as mock_datetime:
        mock_datetime.now.return_value = mock_start_date

        span1 = Span("1", "a", SpanType.BASE, "b", "op1")
        storage.append_span(span1)
        time.sleep(0.1)

        # mock new day
        mock_datetime.now.return_value = mock_start_date + timedelta(minutes=1)

        span2 = Span("2", "c", SpanType.BASE, "d", "op2")
        storage.append_span(span2)
        time.sleep(0.1)

    # origin filename need exists
    assert os.path.exists(storage.filename)

    # get roll over filename
    dated_filename = os.path.join(
        os.path.dirname(storage.filename),
        f"{os.path.basename(storage.filename).split('.')[0]}_2023-10-18.jsonl",
    )

    assert os.path.exists(dated_filename)

    # check origin filename just include the second span
    spans_in_original_file = read_spans_from_file(storage.filename)
    assert len(spans_in_original_file) == 1
    assert spans_in_original_file[0]["trace_id"] == "2"

    # check the roll over filename just include the first span
    spans_in_dated_file = read_spans_from_file(dated_filename)
    assert len(spans_in_dated_file) == 1
    assert spans_in_dated_file[0]["trace_id"] == "1"


@pytest.mark.asyncio
@pytest.mark.parametrize("storage_container", [{"batch_size": 5}], indirect=True)
async def test_container_flush_policy(
    storage_container: SpanStorageContainer, storage: FileSpanStorage
):
    storage_container.append_storage(storage)
    span = Span("1", "a", SpanType.BASE, "b", "op1")

    filename = storage.filename

    for _ in range(storage_container.batch_size - 1):
        storage_container.append_span(span)

    spans_in_file = read_spans_from_file(filename)
    assert len(spans_in_file) == 0

    # Trigger batch write
    storage_container.append_span(span)
    await asyncio.sleep(0.1)

    spans_in_file = read_spans_from_file(filename)
    assert len(spans_in_file) == storage_container.batch_size



import copy
import json
import os
from typing import List, Optional, Tuple, Union

import click

from dbgpt.configs.model_config import ROOT_PATH
from dbgpt.util.console import CliLogger
from dbgpt.util.i18n_utils import _

_DEFAULT_PATH = os.path.join(
    ROOT_PATH, "packages", "dbgpt-serve", "src", "dbgpt_serve", "flow", "compat"
)

clog = CliLogger()


@click.group("flow")
def tool_flow_cli_group():
    """FLow tools."""
    pass


@tool_flow_cli_group.command()
@click.option(
    "--module",
    type=str,
    default=None,
    required=False,
    help=_(
        "The module to scan, if not set, will scan all DB-GPT modules("
        "'dbgpt,dbgpt_client,dbgpt_ext,dbgpt_serve,dbgpt_app')."
    ),
)
@click.option(
    "--output",
    type=str,
    default=None,
    required=False,
    help=_(
        "The output path, if not set, will print to "
        "packages/dbgpt-serve/src/dbgpt_serve/flow/compat/"
    ),
)
@click.option(
    "--curr_version",
    type=str,
    default=None,
    required=False,
    help=_(
        "The current version of the flow, if not set, will read from dbgpt.__version__"
    ),
)
@click.option(
    "--last_support_version",
    type=str,
    default=None,
    required=False,
    help=_(
        "The last version to compatible, if not set, will big than the current version "
        "by one minor version."
    ),
)
@click.option(
    "--update-template",
    is_flag=True,
    default=False,
    required=False,
    help=_("Update the template file."),
)
def gen_compat(
    module: Optional[str],
    output: Optional[str],
    curr_version: Optional[str],
    last_support_version: Optional[str],
    update_template: bool = False,
):
    """Generate the compatibility flow mapping file."""
    from dbgpt.util.cli._module import _scan_awel_flow

    lang = os.getenv("LANGUAGE")

    modules = []
    if module:
        for m in module.split(","):
            modules.append(m.strip())
    if not output:
        output = _DEFAULT_PATH
    if not curr_version:
        from dbgpt import __version__

        curr_version = __version__
    if not last_support_version:
        last_support_version = curr_version.split(".")
        last_support_version[-2] = str(int(last_support_version[-2]) + 1)
        last_support_version[-1] = "x"
        last_support_version = ".".join(last_support_version)

    output_dicts = {
        "curr_version": curr_version,
        "last_support_version": last_support_version,
        "compat": [],
    }
    flows = _scan_awel_flow(modules)
    for flow in flows:
        output_dicts["compat"].append(flow.to_dict())
    os.makedirs(output, exist_ok=True)
    output_file = os.path.join(output, curr_version + "_compat_flow.json")
    user_input = clog.ask(f"Output to {output_file}, do you want to continue?(y/n)")
    save_compat = True
    if not user_input or user_input.lower() != "y":
        clog.info("Cancelled")
        save_compat = False
    if save_compat:
        with open(output_file, "w") as f:
            import json

            json.dump(output_dicts, f, ensure_ascii=False, indent=4)

    if update_template:
        _update_template(output_dicts, lang=lang)


def _update_template(
    compat_data: dict, template_file: Optional[str] = None, lang: str = "en"
):
    from dbgpt.core.awel.dag.base import DAGNode
    from dbgpt.core.awel.flow.base import (
        _TYPE_REGISTRY,
    )
    from dbgpt.core.awel.flow.compat import (
        FlowCompatMetadata,
        _register_flow_compat,
        get_new_class_name,
    )
    from dbgpt.core.awel.flow.flow_factory import (
        ResourceMetadata,
        ViewMetadata,
        fill_flow_panel,
    )
    from dbgpt_serve.flow.api.schemas import ServerResponse
    from dbgpt_serve.flow.service.service import (
        _get_flow_templates_from_files,
        _parse_flow_template_from_json,
    )

    last_support_version = compat_data["last_support_version"]
    curr_version = compat_data["curr_version"]

    for data in compat_data["compat"]:
        metadata = FlowCompatMetadata(**data)
        _register_flow_compat(curr_version, last_support_version, metadata)

    templates: List[Tuple[str, ServerResponse]] = []
    if template_file:
        with open(template_file, "r") as f:
            data = json.load(f)
            templates.append((template_file, _parse_flow_template_from_json(data)))
    else:
        templates.extend(_get_flow_templates_from_files(lang))

    new_templates: List[Tuple[str, ServerResponse]] = []

    def metadata_func(old_metadata: Union[ViewMetadata, ResourceMetadata]):
        type_cls = old_metadata.type_cls
        if type_cls in _TYPE_REGISTRY:
            return None
        new_type_cls = None
        try:
            new_type_cls = get_new_class_name(type_cls)
            if not new_type_cls or new_type_cls not in _TYPE_REGISTRY:
                return None
            obj_type = _TYPE_REGISTRY[new_type_cls]
            if isinstance(old_metadata, ViewMetadata):
                if not isinstance(obj_type, DAGNode):
                    return None
                obj_type.metadata
            elif isinstance(old_metadata, ResourceMetadata):
                metadata_attr = f"_resource_metadata_{obj_type.__name__}"
                return getattr(obj_type, metadata_attr)
            else:
                raise ValueError(f"Unknown metadata type: {type(old_metadata)}")
        except Exception as e:
            clog.warning(
                f"Error get metadata for {type_cls}: {str(e)}, new_type_cls: "
                f"{new_type_cls}"
            )
            return None

    for template_file, template in templates:
        new_flow_template = copy.deepcopy(template)
        try:
            fill_flow_panel(
                new_flow_template,
                metadata_func,
                ignore_options_error=True,
                update_id=True,
            )
            new_templates.append((template_file, new_flow_template))
        except Exception as e:
            import traceback

            traceback.print_exc()
            clog.warning(f"Error fill flow panel for {template_file}: {str(e)}")

    user_input = clog.ask("Do you want to update the template file?(y/n)")
    if not user_input or user_input.lower() != "y":
        clog.info("Cancelled")
        return
    for template_file, flow in new_templates:
        template_dict = {"flow": flow.model_dump()}
        dag_json = json.dumps(template_dict, indent=4, ensure_ascii=False)
        with open(template_file, "w") as f:
            f.write(dag_json)



"""Tool resource pack module."""

import logging
import os
import ssl
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union, cast

from mcp import ClientSession

from dbgpt.util.json_utils import parse_or_raise_error

from ...util.mcp_utils import sse_client
from ..base import EXECUTE_ARGS_TYPE, PARSE_EXECUTE_ARGS_FUNCTION, ResourceType, T
from ..pack import Resource, ResourcePack
from .base import DB_GPT_TOOL_IDENTIFIER, BaseTool, FunctionTool, ToolFunc
from .exceptions import ToolExecutionException, ToolNotFoundException

ToolResourceType = Union[Resource, BaseTool, List[BaseTool], ToolFunc, List[ToolFunc]]

logger = logging.getLogger(__name__)


def _is_function_tool(resources: Any) -> bool:
    return (
        callable(resources)
        and hasattr(resources, DB_GPT_TOOL_IDENTIFIER)
        and getattr(resources, DB_GPT_TOOL_IDENTIFIER)
        and hasattr(resources, "_tool")
        and isinstance(getattr(resources, "_tool"), BaseTool)
    )


def _is_tool(resources: Any) -> bool:
    return isinstance(resources, BaseTool) or _is_function_tool(resources)


def _to_tool_list(
    resources: ToolResourceType, unpack: bool = False, ignore_error: bool = False
) -> List[Resource]:
    def parse_tool(r):
        if isinstance(r, BaseTool):
            return [r]
        elif _is_function_tool(r):
            return [cast(FunctionTool, getattr(r, "_tool"))]
        elif isinstance(r, ResourcePack):
            if not unpack:
                return [r]
            new_list = []
            for p in r.sub_resources:
                new_list.extend(parse_tool(p))
            return new_list
        elif isinstance(r, Sequence):
            new_list = []
            for t in r:
                new_list.extend(parse_tool(t))
            return new_list
        elif ignore_error:
            logger.warning(
                f"_to_tool_list: unknown resource type={type(r).__name__}, "
                f"name={getattr(r, 'name', '?')}, ignored"
            )
            return []
        else:
            raise ValueError("Invalid tool resource type")

    return parse_tool(resources)


def json_parse_execute_args_func(input_str: str) -> Optional[EXECUTE_ARGS_TYPE]:
    """Parse the execute arguments."""
    # The position arguments is empty
    args = ()
    kwargs = parse_or_raise_error(input_str)
    if kwargs is not None and isinstance(kwargs, list) and len(kwargs) == 0:
        kwargs = {}
    return args, kwargs


class ToolPack(ResourcePack):
    """Tool resource pack class."""

    def __init__(
        self, resources: ToolResourceType, name: str = "Tool Resource Pack", **kwargs
    ):
        """Initialize the tool resource pack."""
        tools = cast(List[Resource], _to_tool_list(resources))
        super().__init__(resources=tools, name=name, **kwargs)

    @classmethod
    def from_resource(
        cls: Type[T],
        resource: Optional[Resource],
        expected_type: Optional[ResourceType] = None,
    ) -> List[T]:
        """Create a resource from another resource."""
        if not resource:
            return []
        tools = _to_tool_list(resource, unpack=True, ignore_error=True)
        typed_tools = [cast(BaseTool, t) for t in tools]
        result = [ToolPack(typed_tools)]  # type: ignore
        return result

    def add_command(
        self,
        command_label: str,
        command_name: str,
        args: Optional[Dict[str, Any]] = None,
        function: Optional[Callable] = None,
        parse_execute_args_func: Optional[PARSE_EXECUTE_ARGS_FUNCTION] = None,
        overwrite: bool = False,
    ) -> None:
        """Add a command to the commands.

        Compatible with the Auto-GPT old plugin system.

        Add a command to the commands list with a label, name, and optional arguments.

        Args:
            command_label (str): The label of the command.
            command_name (str): The name of the command.
            args (dict, optional): A dictionary containing argument names and their
              values. Defaults to None.
            function (callable, optional): A callable function to be called when
                the command is executed. Defaults to None.
            parse_execute_args (callable, optional): A callable function to parse the
                execute arguments. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the command if it already
                exists. Defaults to False.
        """
        if args is not None:
            tool_args = {}
            for name, value in args.items():
                if isinstance(value, dict):
                    tool_args[name] = {
                        "name": name,
                        "type": value.get("type", "str"),
                        "description": value.get("description", str(value)),
                        "required": value.get("required", False),
                    }
                    if "title" in value:
                        tool_args[name]["title"] = value["title"]
                    if "default" in value:
                        tool_args[name]["default"] = value["default"]
                else:
                    tool_args[name] = {
                        "name": name,
                        "type": "str",
                        "description": value,
                    }
        else:
            tool_args = {}
        if not function:
            raise ValueError("Function must be provided")

        ft = FunctionTool(
            name=command_name,
            func=function,
            args=tool_args,
            description=command_label,
            parse_execute_args_func=parse_execute_args_func,
        )
        self.append(ft, overwrite=overwrite)

    def _get_execution_tool(
        self,
        name: Optional[str] = None,
    ) -> BaseTool:
        if not name or name not in self._resources:
            raise ToolNotFoundException("No tool found for execution")
        return cast(BaseTool, self._resources[name])

    def _get_call_args(self, arguments: Dict[str, Any], tl: BaseTool) -> Dict[str, Any]:
        """Get the call arguments."""
        # Delete non-defined parameters
        diff_args = list(set(arguments.keys()).difference(set(tl.args.keys())))
        for arg_name in diff_args:
            del arguments[arg_name]
        return arguments

    def parse_execute_args(
        self, resource_name: Optional[str] = None, input_str: Optional[str] = None
    ) -> Optional[EXECUTE_ARGS_TYPE]:
        """Parse the execute arguments."""
        try:
            tl = self._get_execution_tool(resource_name)
            return tl.parse_execute_args(input_str=input_str)
        except ToolNotFoundException:
            return None

    def execute(
        self,
        *args,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> Any:
        """Execute the tool.

        Args:
            *args: The positional arguments.
            resource_name (str, optional): The tool name to be executed.
            **kwargs: The keyword arguments.

        Returns:
            Any: The result of the tool execution.
        """
        tl = self._get_execution_tool(resource_name)
        try:
            arguments = {k: v for k, v in kwargs.items()}
            arguments = self._get_call_args(arguments, tl)
            if tl.is_async:
                raise ToolExecutionException("Async execution is not supported")
            else:
                return tl.execute(**arguments)
        except Exception as e:
            raise ToolExecutionException(f"Execution error: {str(e)}")

    async def async_execute(
        self,
        *args,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> Any:
        """Execute the tool asynchronously.

        Args:
            *args: The positional arguments.
            resource_name (str, optional): The tool name to be executed.
            **kwargs: The keyword arguments.

        Returns:
            Any: The result of the tool execution.
        """
        tl = self._get_execution_tool(resource_name)
        try:
            arguments = {k: v for k, v in kwargs.items()}
            arguments = self._get_call_args(arguments, tl)
            if tl.is_async:
                return await tl.async_execute(**arguments)
            else:
                # TODO: Execute in a separate executor
                return tl.execute(**arguments)
        except Exception as e:
            raise ToolExecutionException(f"Execution error: {str(e)}")

    def is_terminal(self, resource_name: Optional[str] = None) -> bool:
        """Check if the tool is terminal."""
        from ...expand.actions.react_action import Terminate

        if not resource_name:
            return False
        tl = self._get_execution_tool(resource_name)
        return isinstance(tl, Terminate)


class AutoGPTPluginToolPack(ToolPack):
    """Auto-GPT plugin tool pack class."""

    def __init__(self, plugin_path: Union[str, List[str]], **kwargs):
        """Create an Auto-GPT plugin tool pack."""
        super().__init__([], **kwargs)
        self._plugin_path = plugin_path
        self._loaded = False

    async def preload_resource(self):
        """Preload the resource."""
        from .autogpt.plugins_util import scan_plugin_file, scan_plugins

        if self._loaded:
            return
        paths = (
            [self._plugin_path]
            if isinstance(self._plugin_path, str)
            else self._plugin_path
        )
        plugins = []
        for path in paths:
            if os.path.isabs(path):
                if not os.path.exists(path):
                    raise ValueError(f"Wrong plugin path configured {path}!")
                if os.path.isfile(path):
                    plugins.extend(scan_plugin_file(path))
                else:
                    plugins.extend(scan_plugins(path))
        for plugin in plugins:
            if not plugin.can_handle_post_prompt():
                continue
            plugin.post_prompt(self)
        self._loaded = True


class MCPToolPack(ToolPack):
    """MCP tool pack class.

    Wrap the MCP SSE server as a tool pack.

    Example:
        .. code-block:: python

            tools = MCPToolPack("http://127.0.0.1:8000/sse")

        If you want to pass the token to the server, you can use the headers parameter:
        .. code-block:: python

            tools = MCPToolPack(
                "http://127.0.0.1:8000/sse"
                default_headers={"Authorization": "Bearer your_token"}
            )
            # Set the default headers for ech server
            tools2 = MCPToolPack(
                "http://127.0.0.1:8000/sse"
                headers = {
                    "http://127.0.0.1:8000/sse": {
                        "Authorization": "Bearer your_token"
                    }
                }
            )

        If you want to set the ssl verify, you can use the ssl_verify parameter:
        .. code-block:: python

            # Default ssl_verify is True
            tools = MCPToolPack(
                "https://your_ssl_domain/sse",
            )

            # Set the default ssl_verify to False to disable ssl verify
            tools2 = MCPToolPack(
                "https://your_ssl_domain/sse", default_ssl_verify=False
            )

            # With Custom CA file
            tools3 = MCPToolPack(
                "https://your_ssl_domain/sse", default_ssl_cafile="/path/to/your/ca.crt"
            )

            # Set the ssl_verify for each server
            import ssl

            tools4 = MCPToolPack(
                "https://your_ssl_domain/sse",
                ssl_verify={
                    "https://your_ssl_domain/sse": ssl.create_default_context(
                        cafile="/path/to/your/ca.crt"
                    ),
                },
            )

    """

    def __init__(
        self,
        mcp_servers: Union[str, List[str]],
        headers: Optional[Dict[str, Dict[str, Any]]] = None,
        default_headers: Optional[Dict[str, Any]] = None,
        ssl_verify: Optional[Dict[str, Union[ssl.SSLContext, str, bool]]] = None,
        default_ssl_verify: Union[ssl.SSLContext, str, bool] = True,
        default_ssl_cafile: Optional[str] = None,
        overwrite_same_tool: bool = True,
        **kwargs,
    ):
        """Create an Auto-GPT plugin tool pack."""
        super().__init__([], **kwargs)
        self._mcp_servers = mcp_servers
        self._loaded = False
        self.tool_server_map = {}
        self._default_headers = default_headers or {}
        self._headers_map = headers or {}
        self.server_headers_map = {}
        if default_ssl_cafile and not ssl_verify and default_ssl_verify:
            default_ssl_verify = ssl.create_default_context(cafile=default_ssl_cafile)

        self._default_ssl_verify = default_ssl_verify
        self._ssl_verify_map = ssl_verify or {}
        self.server_ssl_verify_map = {}
        self._overwrite_same_tool = overwrite_same_tool

    def switch_mcp_input_schema(self, input_schema: dict):
        args = {}
        try:
            properties = input_schema["properties"]
            required = input_schema.get("required", [])
            for k, v in properties.items():
                arg = {}

                title = v.get("title", None)
                description = v.get("description", None)
                items = v.get("items", None)
                items_str = str(items) if items else None
                any_of = v.get("anyOf", None)
                any_of_str = str(any_of) if any_of else None

                default = v.get("default", None)
                type = v.get("type", "string")

                arg["type"] = type
                if title:
                    arg["title"] = title
                arg["description"] = description or items_str or any_of_str or str(v)
                arg["required"] = True if k in required else False
                if default:
                    arg["default"] = default
                args[k] = arg
            return args
        except Exception as e:
            raise ValueError(f"MCP input_schema can't parase!{str(e)},{input_schema}")

    async def preload_resource(self):
        """Preload the resource."""
        server_list = []
        if isinstance(self._mcp_servers, List):
            server_list = self._mcp_servers.copy()
        else:
            server_list = self._mcp_servers.split(";")

        for server in server_list:
            server_headers = self._headers_map.get(server, self._default_headers)
            self.server_headers_map[server] = server_headers
            server_ssl_verify = self._ssl_verify_map.get(
                server, self._default_ssl_verify
            )
            self.server_ssl_verify_map[server] = server_ssl_verify

            async with sse_client(
                url=server, headers=server_headers, verify=server_ssl_verify
            ) as (read, write):
                async with ClientSession(read, write) as session:
                    # Initialize the connection
                    await session.initialize()
                    result = await session.list_tools()
                    for tool in result.tools:
                        tool_name = tool.name
                        self.tool_server_map[tool_name] = server
                        args = self.switch_mcp_input_schema(tool.inputSchema)

                        async def call_mcp_tool(
                            tool_name=tool_name, server=server, **kwargs
                        ):
                            try:
                                headers_to_use = self.server_headers_map.get(server, {})
                                ssl_verify_to_use = self.server_ssl_verify_map.get(
                                    server, True
                                )
                                async with sse_client(
                                    url=server,
                                    headers=headers_to_use,
                                    verify=ssl_verify_to_use,
                                ) as (read, write):
                                    async with ClientSession(read, write) as session:
                                        # Initialize the connection
                                        await session.initialize()
                                        return await session.call_tool(
                                            tool_name, arguments=kwargs
                                        )
                            except Exception as e:
                                raise ValueError(f"MCP Call Exception! {str(e)}")

                        self.add_command(
                            tool.description,
                            tool_name,
                            args,
                            call_mcp_tool,
                            parse_execute_args_func=json_parse_execute_args_func,
                            overwrite=self._overwrite_same_tool,
                        )
        self._loaded = True



import asyncio
import logging
from enum import Enum, auto
from typing import Dict, Optional, cast

from lyric import CodeResult, DefaultLyricDriver, PyTaskResourceConfig

from dbgpt.component import BaseComponent, SystemApp

logger = logging.getLogger(__name__)


class ServerState(Enum):
    INIT = auto()
    STARTING = auto()
    READY = auto()
    STOPPING = auto()
    STOPPED = auto()


class CodeServer(BaseComponent):
    def __init__(self, system_app: Optional[SystemApp] = None):
        self.sys_app = system_app
        super().__init__(system_app)
        self._lcd = DefaultLyricDriver()
        self._init_lock = asyncio.Lock()
        self._state = ServerState.INIT
        self._ready_event = asyncio.Event()

    def init_app(self, system_app: SystemApp):
        self.sys_app = system_app

    def before_start(self):
        if self._state == ServerState.INIT:
            self._state = ServerState.STARTING
            self._lcd.start()

    def before_stop(self):
        self._state = ServerState.STOPPING
        self._lcd.stop()
        self._state = ServerState.STOPPED
        self._ready_event.clear()

    async def async_after_start(self):
        await self._ensure_initialized()

    async def _ensure_initialized(self):
        """Ensure the server is initialized and workers are loaded."""
        if self._state == ServerState.READY:
            return

        async with self._init_lock:
            # Double check after acquiring lock
            if self._state == ServerState.READY:
                return

            if self._state == ServerState.INIT:
                logger.info("Starting code server...")
                self._state = ServerState.STARTING
                self._lcd.start()
            if self._state == ServerState.STARTING:
                await self._lcd.lyric.load_default_workers()
                self._state = ServerState.READY
                self._ready_event.set()
                logger.info("Code server is ready")

    async def wait_ready(self, timeout: Optional[float] = None) -> bool:
        """Wait until server is ready.

        Args:
            timeout: Maximum time to wait in seconds. None means wait forever.

        Returns:
            bool: True if server is ready, False if timeout occurred
        """
        try:
            await asyncio.wait_for(self._ready_event.wait(), timeout)
            return True
        except asyncio.TimeoutError:
            return False

    async def exec(
        self, code: str, lang: str, resources: Optional[PyTaskResourceConfig] = None
    ) -> CodeResult:
        await self._ensure_initialized()
        return await self._lcd.exec(code, lang, resources=resources)

    async def exec1(
        self,
        code: str,
        input_bytes: bytes,
        call_name: str,
        lang: str = "python",
        resources: Optional[PyTaskResourceConfig] = None,
    ) -> CodeResult:
        await self._ensure_initialized()
        return await self._lcd.exec1(
            code, input_bytes, call_name, lang=lang, resources=resources
        )

    async def parse_awel(self, code: str) -> Optional[Dict]:
        """Parse the AWEL code.

        Return the flow metadata.
        """
        raise NotImplementedError

    async def run_awel_operator(self, code: str):
        """Run an AWEL operator in remote mode."""
        raise NotImplementedError


_SYSTEM_APP: Optional[SystemApp] = None


def initialize_code_server(system_app: SystemApp):
    """Initialize the code server."""
    global _SYSTEM_APP
    _SYSTEM_APP = system_app
    code_server = CodeServer(system_app)
    system_app.register_instance(code_server)


async def get_code_server(
    system_app: Optional[SystemApp] = None,
    wait_ready: bool = True,
    timeout: Optional[float] = None,
) -> CodeServer:
    """Return the code server.

    Args:
        system_app (Optional[SystemApp]): The system app. Defaults to None.
        wait_ready (bool): Whether to wait for server to be ready. Defaults to True.
        timeout (Optional[float]): Maximum time to wait in seconds. None means wait
            forever.

    Returns:
        CodeServer: The code server.
    """
    if not _SYSTEM_APP:
        if not system_app:
            system_app = SystemApp()
        initialize_code_server(system_app)

    app = system_app or _SYSTEM_APP
    server = CodeServer.get_instance(cast(SystemApp, app))

    if wait_ready:
        await server._ensure_initialized()
        if not await server.wait_ready(timeout):
            raise TimeoutError("Timeout waiting for code server to be ready")

    return server



"""Chart visualization protocol conversion class."""

import json
from typing import Any, Dict, Optional

from ..base import Vis


def default_chart_type_prompt() -> str:
    """Return prompt information for the default chart type.

    This function is moved from excel_analyze/chat.py,and used by subclass.

    Returns:
        str: prompt information for the default chart type.
    """
    antv_charts = [
        {"response_line_chart": "used to display comparative trend analysis data"},
        {
            "response_pie_chart": "suitable for scenarios such as proportion and "
            "distribution statistics"
        },
        {
            "response_table": "suitable for display with many display columns or "
            "non-numeric columns"
        },
        {
            "response_scatter_chart": "Suitable for exploring relationships between "
            "variables, detecting outliers, etc."
        },
        {
            "response_bubble_chart": "Suitable for relationships between multiple "
            "variables, highlighting outliers or special situations, etc."
        },
        {
            "response_donut_chart": "Suitable for hierarchical structure representation"
            ", category proportion display and highlighting key categories, etc."
        },
        {
            "response_area_chart": "Suitable for visualization of time series data, "
            "comparison of multiple groups of data, analysis of data change trends, "
            "etc."
        },
        {
            "response_heatmap": "Suitable for visual analysis of time series data, "
            "large-scale data sets, distribution of classified data, etc."
        },
    ]
    return "\n".join(
        f"{key}:{value}"
        for dict_item in antv_charts
        for key, value in dict_item.items()
    )


class VisChart(Vis):
    """Chart visualization protocol conversion class."""

    def render_prompt(self) -> Optional[str]:
        """Return the prompt for the vis protocol."""
        return default_chart_type_prompt()

    def sync_generate_param(self, **kwargs) -> Optional[Dict[str, Any]]:
        """Generate the parameters required by the vis protocol."""
        chart = kwargs.get("chart", None)
        data_df = kwargs.get("data_df", None)

        if not chart:
            raise ValueError(
                f"Parameter information is missing and {self.vis_tag} protocol "
                "conversion cannot be performed."
            )

        sql = chart.get("sql", None)
        param = {}
        if not sql or len(sql) <= 0:
            return None

        param["sql"] = sql
        param["type"] = chart.get("display_type", "response_table")
        param["title"] = chart.get("title", "")
        param["describe"] = chart.get("thought", "")

        param["data"] = json.loads(
            data_df.to_json(orient="records", date_format="iso", date_unit="s")
        )
        return param

    @classmethod
    def vis_tag(cls) -> str:
        """Return the tag name of the vis protocol."""
        return "vis-db-chart"



import functools
import shutil
import subprocess
from pathlib import Path

import click

from ..console import CliLogger
from .base import DEFAULT_PACKAGE_TYPES

cl = CliLogger()


def check_build_tools_installed():
    """Check if any supported build tools are installed (uv, poetry, build, or
    setuptools)

    Warns if uv is not installed but does not exit.
    Only exits if no build tools are available.
    """
    tools_available = []

    # Check for uv (preferred tool)
    if shutil.which("uv"):
        try:
            subprocess.run(
                ["uv", "--version"],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            tools_available.append("uv")
        except subprocess.CalledProcessError:
            pass

    # Check for poetry
    if shutil.which("poetry"):
        try:
            subprocess.run(
                ["poetry", "--version"],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            tools_available.append("poetry")
        except subprocess.CalledProcessError:
            pass

    # Check for build package
    try:
        # Check if python and build are available
        result = subprocess.run(
            ["python", "-c", "import build; print('yes')"],
            check=True,
            capture_output=True,
            text=True,
        )
        if "yes" in result.stdout:
            tools_available.append("build")
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass

    # Check for python with setuptools
    try:
        # Check if python and setuptools are available for basic package building
        result = subprocess.run(
            ["python", "-c", "import setuptools; print('yes')"],
            check=True,
            capture_output=True,
            text=True,
        )
        if "yes" in result.stdout:
            tools_available.append("setuptools")
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass

    # Provide appropriate feedback based on available tools
    if not tools_available:
        cl.error("No build tools found. Please install one of the following:")
        cl.error(" - uv: A fast Python package installer and resolver (recommended)")
        cl.error(" - poetry: A Python package and dependency manager")
        cl.error(" - build: A PEP 517 compatible Python package builder")
        cl.error(" - setuptools: A classic Python package build system")
        cl.error(
            "For uv: https://github.com/astral-sh/uv\n"
            "For poetry: https://python-poetry.org/docs/#installation\n"
            "For build: pip install build\n"
            "For setuptools: pip install setuptools",
            exit_code=1,
        )
    elif "uv" not in tools_available:
        cl.warning(
            "uv is not installed. We recommend using uv for better performance."
            " Install with: 'pip install uv' or visit https://github.com/astral-sh/uv"
        )
        if "build" not in tools_available:
            cl.warning(
                "build package is not installed. For better compatibility without uv,"
                " we recommend installing it with: 'pip install build'"
            )

    return tools_available


def add_tap_options(func):
    @click.option(
        "-r",
        "--repo",
        type=str,
        default=None,
        required=False,
        help="The repository to install the dbgpts from",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def add_add_common_options(func):
    @click.option(
        "-r",
        "--repo",
        type=str,
        default=None,
        required=False,
        help="The repository to install the dbgpts from",
    )
    @click.option(
        "-U",
        "--update",
        type=bool,
        required=False,
        default=False,
        is_flag=True,
        help="Whether to update the repo",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.command(name="install")
@add_add_common_options
@click.argument("names", type=str, nargs=-1)
def install(repo: str | None, update: bool, names: list[str]):
    """Install your dbgpts(operators,agents,workflows or apps)"""
    from .repo import _install_default_repos_if_no_repos, install

    check_build_tools_installed()
    _install_default_repos_if_no_repos()
    for name in names:
        install(name, repo, with_update=update)


@click.command(name="uninstall")
@click.argument("names", type=str, nargs=-1)
def uninstall(names: list[str]):
    """Uninstall your dbgpts(operators,agents,workflows or apps)"""
    from .repo import uninstall

    for name in names:
        uninstall(name)


@click.command(name="reinstall")
@add_add_common_options
@click.argument("names", type=str, nargs=-1)
def reinstall(repo: str | None, update: bool, names: list[str]):
    """Reinstall your dbgpts(operators,agents,workflows or apps)"""
    from .repo import reinstall

    for name in names:
        reinstall(name, repo, with_update=update)


@click.command(name="list-remote")
@add_add_common_options
def list_all_apps(
    repo: str | None,
    update: bool,
):
    """List all available dbgpts"""
    from .repo import _install_default_repos_if_no_repos, list_repo_apps

    _install_default_repos_if_no_repos()
    list_repo_apps(repo, with_update=update)


@click.command(name="list")
def list_installed_apps():
    """List all installed dbgpts"""
    from .repo import list_installed_apps

    list_installed_apps()


@click.command(name="list")
def list_repos():
    """List all repos"""
    from .repo import _print_repos

    _print_repos()


@click.command(name="add")
@add_tap_options
@click.option(
    "-b",
    "--branch",
    type=str,
    default=None,
    required=False,
    help="The branch of the repository(Just for git repo)",
)
@click.option(
    "--url",
    type=str,
    required=True,
    help="The URL of the repo",
)
def add_repo(repo: str, branch: str | None, url: str):
    """Add a new repo"""
    from .repo import add_repo

    add_repo(repo, url, branch)


@click.command(name="remove")
@click.argument("repo", type=str)
def remove_repo(repo: str):
    """Remove the specified repo"""
    from .repo import remove_repo

    remove_repo(repo)


@click.command(name="update")
@click.option(
    "-r",
    "--repo",
    type=str,
    default=None,
    required=False,
    help="The repository to update(Default: all repos)",
)
def update_repo(repo: str | None):
    """Update the specified repo"""
    from .repo import list_repos, update_repo

    for p in list_repos():
        if repo:
            if p == repo or repo == "all":
                print(f"Updating repo '{p}'...")
                update_repo(p)

        else:
            print(f"Updating repo '{p}'...")
            update_repo(p)


@click.command(name="app")
@click.option(
    "-n",
    "--name",
    type=str,
    required=True,
    help="The name you want to give to the dbgpt",
)
@click.option(
    "-l",
    "--label",
    type=str,
    default=None,
    required=False,
    help="The label of the dbgpt",
)
@click.option(
    "-d",
    "--description",
    type=str,
    default=None,
    required=False,
    help="The description of the dbgpt",
)
@click.option(
    "-t",
    "--type",
    type=click.Choice(DEFAULT_PACKAGE_TYPES),
    default="flow",
    required=False,
    help="The type of the dbgpt",
)
@click.option(
    "--definition_type",
    type=click.Choice(["json", "python"]),
    default="json",
    required=False,
    help="The definition type of the dbgpt",
)
@click.option(
    "-C",
    "--directory",
    type=str,
    default=None,
    required=False,
    help="The working directory of the dbgpt(defaults to the current directory).",
)
def new_dbgpts(
    name: str,
    label: str | None,
    description: str | None,
    type: str,
    definition_type: str,
    directory: str | None,
):
    """New a dbgpts module structure"""
    if not label:
        # Set label to the name
        default_label = name.replace("-", " ").replace("_", " ").title()
        label = click.prompt(
            "Please input the label of the dbgpt", default=default_label
        )
    if not description:
        # Read with click
        description = click.prompt(
            "Please input the description of the dbgpt", default=""
        )
    if not directory:
        # Set directory to the current directory(abs path)
        directory = click.prompt(
            "Please input the working directory of the dbgpt",
            default=str(Path.cwd()),
            type=click.Path(exists=True, file_okay=False, dir_okay=True),
        )

    check_build_tools_installed()
    from .template import create_template

    create_template(name, label, description, type, definition_type, directory)



"""Knowledge resource."""

import dataclasses
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

import cachetools

from dbgpt._private.config import Config
from dbgpt.core import Chunk
from dbgpt.rag.embedding.embedding_factory import RerankEmbeddingFactory
from dbgpt.rag.retriever.rerank import RerankEmbeddingsRanker
from dbgpt.util.cache_utils import cached

from .base import Resource, ResourceParameters, ResourceType

if TYPE_CHECKING:
    from dbgpt.rag.retriever.base import BaseRetriever
    from dbgpt.storage.vector_store.filters import MetadataFilters

CFG = Config()


@dataclasses.dataclass
class RetrieverResourceParameters(ResourceParameters):
    """Retriever resource parameters."""

    pass


class RetrieverResource(Resource[ResourceParameters]):
    """Retriever resource.

    Retrieve knowledge chunks from a retriever.
    """

    def __init__(self, name: str, retriever: "BaseRetriever"):
        """Create a new RetrieverResource."""
        self._name = name
        self._retriever = retriever
        app_config = CFG.SYSTEM_APP.config.configs.get("app_config")
        self.need_rerank = bool(app_config.models.rerankers)
        if self.need_rerank:
            rerank_embeddings = RerankEmbeddingFactory.get_instance(
                CFG.SYSTEM_APP
            ).create()
            self.reranker = RerankEmbeddingsRanker(
                rerank_embeddings, topk=app_config.rag.rerank_top_k
            )
        else:
            self.reranker = None

    @property
    def name(self) -> str:
        """Return the resource name."""
        return self._name

    @property
    def retriever_name(self) -> str:
        """Return the resource name."""
        return ""

    @property
    def retriever_desc(self) -> str:
        """Return the retriever desc."""
        return ""

    @property
    def retriever(self) -> "BaseRetriever":
        """Return the retriever."""
        return self._retriever

    @classmethod
    def type(cls) -> ResourceType:
        """Return the resource type."""
        return ResourceType.Knowledge

    @classmethod
    def resource_parameters_class(cls, **kwargs) -> Type[ResourceParameters]:
        """Return the resource parameters class."""
        return RetrieverResourceParameters

    @cached(cachetools.TTLCache(maxsize=100, ttl=10))
    async def get_prompt(
        self,
        *,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> Tuple[str, Optional[Dict]]:
        """Get the prompt for the resource."""
        if not question:
            raise ValueError("Question is required for knowledge resource.")
        chunks = await self.retrieve(question)
        if self.need_rerank and len(chunks) > 1:
            chunks = self.reranker.rank(candidates_with_scores=chunks, query=question)

        content = "\n".join(
            [f"--{i}--:" + chunk.content for i, chunk in enumerate(chunks)]
        )
        prompt_template = f"\nResources-{self.name}:\n {content}"
        prompt_template_zh = f"\n资源-{self.name}:\n {content}"
        if lang == "en":
            return prompt_template, self._get_references(chunks)
        return prompt_template_zh, self._get_references(chunks)

    async def get_resources(
        self,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
    ) -> Tuple[Optional[List[Chunk]], str, Optional[Dict]]:
        """Get the chunks for the resource."""
        if not question:
            raise ValueError("Question is required for knowledge resource.")
        chunks = await self.retrieve(question)
        if self.need_rerank and len(chunks) > 1:
            chunks = self.reranker.rank(candidates_with_scores=chunks, query=question)

        prompt_template = """Resources-{name}:\n {content}"""
        prompt_template_zh = """资源-{name}:\n {content}"""
        if lang == "en":
            return chunks, prompt_template, self._get_references(chunks)
        else:
            return chunks, prompt_template_zh, self._get_references(chunks)

    def _get_references(self, docs: List[Chunk]) -> Optional[Dict]:
        references_dict = {}
        for chunk in docs:
            doc_name = None
            prop_field = chunk.metadata.get("prop_field", None)
            if prop_field:
                doc_name = prop_field.get("source", None)
            if not doc_name:
                doc_name = chunk.metadata.get("metadata", "-")
            if doc_name not in references_dict:
                references_dict[doc_name] = {
                    "name": doc_name,
                    "chunks": [
                        {
                            "id": chunk.chunk_id,
                            "content": chunk.content,
                            "meta_info": doc_name,
                            "recall_score": chunk.score,
                            "retriever": chunk.retriever,
                        }
                    ],
                }
            else:
                references_dict[doc_name]["chunks"].append(
                    {
                        "id": chunk.chunk_id,
                        "content": chunk.content,
                        "meta_info": doc_name,
                        "recall_score": chunk.score,
                        "retriever": chunk.retriever,
                    }
                )
        return {self.type().value: list(references_dict.values())}

    async def async_execute(
        self, *args, resource_name: Optional[str] = None, **kwargs
    ) -> Any:
        """Execute the resource asynchronously."""
        return await self.retrieve(*args, **kwargs)

    async def retrieve(
        self,
        query: str,
        filters: Optional["MetadataFilters"] = None,
        score: float = 0.0,
    ) -> List["Chunk"]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text.
            filters: (Optional[MetadataFilters]) metadata filters.
            score: (float) similarity score.

        Returns:
            List[Chunk]: list of chunks
        """
        return await self.retriever.aretrieve_with_scores(query, score, filters)



from dbgpt.model.cluster.apiserver.api import run_apiserver
from dbgpt.model.cluster.base import (
    EmbeddingsRequest,
    PromptRequest,
    WorkerApplyRequest,
    WorkerParameterRequest,
    WorkerStartupRequest,
)
from dbgpt.model.cluster.controller.controller import (
    BaseModelController,
    ModelRegistryClient,
    run_model_controller,
)
from dbgpt.model.cluster.manager_base import WorkerManager, WorkerManagerFactory
from dbgpt.model.cluster.registry import ModelRegistry
from dbgpt.model.cluster.worker.default_worker import DefaultModelWorker
from dbgpt.model.cluster.worker.manager import (
    initialize_worker_manager_in_client,
    run_worker_manager,
    worker_manager,
)
from dbgpt.model.cluster.worker.remote_manager import RemoteWorkerManager
from dbgpt.model.cluster.worker_base import ModelWorker

__all__ = [
    "EmbeddingsRequest",
    "PromptRequest",
    "WorkerApplyRequest",
    "WorkerParameterRequest",
    "WorkerStartupRequest",
    "WorkerManager",
    "WorkerManagerFactory",
    "ModelWorker",
    "DefaultModelWorker",
    "worker_manager",
    "run_worker_manager",
    "initialize_worker_manager_in_client",
    "ModelRegistry",
    "BaseModelController",
    "ModelRegistryClient",
    "RemoteWorkerManager",
    "run_model_controller",
    "run_apiserver",
]



import logging
import os
from concurrent.futures import Executor
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Dict,
    List,
    Optional,
    Type,
    Union,
    cast,
)

from dbgpt.core import MessageConverter, ModelMetadata, ModelOutput, ModelRequest
from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.model.proxy.base import (
    AsyncGenerateStreamFunction,
    GenerateStreamFunction,
    ProxyLLMClient,
    ProxyTokenizer,
    TiktokenProxyTokenizer,
    register_proxy_model_adapter,
)
from dbgpt.model.proxy.llms.chatgpt import OpenAICompatibleDeployModelParameters
from dbgpt.model.proxy.llms.proxy_model import ProxyModel, parse_model_request
from dbgpt.util.i18n_utils import _

if TYPE_CHECKING:
    from anthropic import AsyncAnthropic, ProxiesTypes

logger = logging.getLogger(__name__)


@auto_register_resource(
    label=_("Claude Proxy LLM"),
    category=ResourceCategory.LLM_CLIENT,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("Claude Proxy LLM"),
    documentation_url="https://docs.anthropic.com/en/api/getting-started",
    show_in_ui=False,
)
@dataclass
class ClaudeDeployModelParameters(OpenAICompatibleDeployModelParameters):
    """Deploy model parameters for Claude."""

    provider: str = "proxy/claude"

    api_base: Optional[str] = field(
        default="${env:ANTHROPIC_BASE_URL:-https://api.anthropic.com}",
        metadata={
            "help": _("The base url of the claude API."),
        },
    )

    api_key: Optional[str] = field(
        default="${env:ANTHROPIC_API_KEY}",
        metadata={
            "help": _("The API key of the claude API."),
            "tags": "privacy",
        },
    )


async def claude_generate_stream(
    model: ProxyModel,
    tokenizer: Any,
    params: Dict[str, Any],
    device: str,
    context_len=2048,
) -> AsyncIterator[ModelOutput]:
    client: ClaudeLLMClient = cast(ClaudeLLMClient, model.proxy_llm_client)
    request = parse_model_request(params, client.default_model, stream=True)
    async for r in client.generate_stream(request):
        yield r


class ClaudeLLMClient(ProxyLLMClient):
    def __init__(
        self,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        model: Optional[str] = None,
        proxies: Optional["ProxiesTypes"] = None,
        timeout: Optional[int] = 240,
        model_alias: Optional[str] = "claude-3-5-sonnet-20241022",
        context_length: Optional[int] = 8192,
        client: Optional["AsyncAnthropic"] = None,
        claude_kwargs: Optional[Dict[str, Any]] = None,
        proxy_tokenizer: Optional[ProxyTokenizer] = None,
    ):
        try:
            import anthropic  # noqa: F401
        except ImportError as exc:
            raise ValueError(
                "Could not import python package: anthropic "
                "Please install anthropic by command `pip install anthropic"
            ) from exc
        if not model:
            model = "claude-3-5-sonnet-20241022"
        self._client = client
        self._model = model
        self._api_key = self._resolve_env_vars(api_key)
        api_base or os.environ.get("ANTHROPIC_BASE_URL", "https://api.anthropic.com")
        self._api_base = self._resolve_env_vars(self._api_base)
        self._proxies = proxies
        self._timeout = timeout
        self._claude_kwargs = claude_kwargs or {}
        self._model_alias = model_alias
        self._proxy_tokenizer = proxy_tokenizer

        super().__init__(
            model_names=[model_alias],
            context_length=context_length,
            proxy_tokenizer=proxy_tokenizer,
        )

    @classmethod
    def new_client(
        cls,
        model_params: ClaudeDeployModelParameters,
        default_executor: Optional[Executor] = None,
    ) -> "ClaudeLLMClient":
        return cls(
            api_key=model_params.api_key,
            api_base=model_params.api_base,
            model=model_params.real_provider_model_name,
            proxies=model_params.http_proxy,
            model_alias=model_params.real_provider_model_name,
            context_length=max(model_params.context_length or 8192, 8192),
        )

    @classmethod
    def param_class(cls) -> Type[ClaudeDeployModelParameters]:
        """Get the model parameters class."""
        return ClaudeDeployModelParameters

    @classmethod
    def generate_stream_function(
        cls,
    ) -> Optional[Union[GenerateStreamFunction, AsyncGenerateStreamFunction]]:
        """Get generate stream function.

        Returns:
            Optional[Union[GenerateStreamFunction, AsyncGenerateStreamFunction]]:
                generate stream function
        """
        return claude_generate_stream

    @property
    def client(self) -> "AsyncAnthropic":
        from anthropic import AsyncAnthropic

        if self._client is None:
            self._client = AsyncAnthropic(
                api_key=self._api_key,
                base_url=self._api_base,
                proxies=self._proxies,
                timeout=self._timeout,
            )
        return self._client

    @property
    def proxy_tokenizer(self) -> ProxyTokenizer:
        if not self._proxy_tokenizer:
            self._proxy_tokenizer = ClaudeProxyTokenizer(self.client)
        return self._proxy_tokenizer

    @property
    def default_model(self) -> str:
        """Default model name"""
        model = self._model
        if not model:
            model = "claude-3-5-sonnet-20241022"
        return model

    def _build_request(
        self, request: ModelRequest, stream: Optional[bool] = False
    ) -> Dict[str, Any]:
        payload = {"stream": stream}
        model = request.model or self.default_model
        payload["model"] = model
        # Apply claude kwargs
        for k, v in self._claude_kwargs.items():
            payload[k] = v
        if request.temperature:
            payload["temperature"] = request.temperature
        if request.max_new_tokens:
            payload["max_tokens"] = request.max_new_tokens
        if request.stop:
            payload["stop"] = request.stop
        if request.top_p:
            payload["top_p"] = request.top_p
        return payload

    async def generate(
        self,
        request: ModelRequest,
        message_converter: Optional[MessageConverter] = None,
    ) -> ModelOutput:
        request = self.local_covert_message(request, message_converter)
        messages, system_messages = request.split_messages()
        payload = self._build_request(request)
        logger.info(
            f"Send request to claude, payload: {payload}\n\n messages:\n{messages}"
        )
        try:
            if len(system_messages) > 1:
                raise ValueError("Claude only supports single system message")
            if system_messages:
                payload["system"] = system_messages[0]
            if "max_tokens" not in payload:
                max_tokens = 1024
            else:
                max_tokens = payload["max_tokens"]
                del payload["max_tokens"]
            response = await self.client.messages.create(
                max_tokens=max_tokens,
                messages=messages,
                **payload,
            )
            usage = None
            finish_reason = response.stop_reason
            if response.usage:
                usage = {
                    "prompt_tokens": response.usage.input_tokens,
                    "completion_tokens": response.usage.output_tokens,
                }
            response_content = response.content
            if not response_content:
                raise ValueError("Response content is empty")
            return ModelOutput(
                text=response_content[0].text,
                error_code=0,
                finish_reason=finish_reason,
                usage=usage,
            )
        except Exception as e:
            return ModelOutput(
                text=f"**Claude Generate Error, Please CheckErrorInfo.**: {e}",
                error_code=1,
            )

    async def generate_stream(
        self,
        request: ModelRequest,
        message_converter: Optional[MessageConverter] = None,
    ) -> AsyncIterator[ModelOutput]:
        request = self.local_covert_message(request, message_converter)
        messages, system_messages = request.split_messages()
        payload = self._build_request(request, stream=True)
        logger.info(
            f"Send request to claude, payload: {payload}\n\n messages:\n{messages}"
        )
        try:
            if len(system_messages) > 1:
                raise ValueError("Claude only supports single system message")
            if system_messages:
                payload["system"] = system_messages[0]
            if "max_tokens" not in payload:
                max_tokens = 1024
            else:
                max_tokens = payload["max_tokens"]
                del payload["max_tokens"]
            if "stream" in payload:
                del payload["stream"]
            full_text = ""
            async with self.client.messages.stream(
                max_tokens=max_tokens,
                messages=messages,
                **payload,
            ) as stream:
                async for text in stream.text_stream:
                    full_text += text
                    raw_usage = stream.current_message_snapshot.usage
                    usage = {
                        "prompt_tokens": raw_usage.input_tokens,
                        "completion_tokens": raw_usage.output_tokens,
                    }
                    yield ModelOutput(text=full_text, error_code=0, usage=usage)
        except Exception as e:
            yield ModelOutput(
                text=f"**Claude Generate Stream Error, Please CheckErrorInfo.**: {e}",
                error_code=1,
            )

    async def models(self) -> List[ModelMetadata]:
        model_metadata = ModelMetadata(
            model=self._model_alias,
            context_length=await self.get_context_length(),
        )
        return [model_metadata]

    async def get_context_length(self) -> int:
        """Get the context length of the model.

        Returns:
            int: The context length.
        # TODO: This is a temporary solution. We should have a better way to get the
            context length.
            eg. get real context length from the openai api.
        """
        return self.context_length


class ClaudeProxyTokenizer(ProxyTokenizer):
    def __init__(self, client: "AsyncAnthropic", concurrency_limit: int = 10):
        self.client = client
        self.concurrency_limit = concurrency_limit
        self._tiktoken_tokenizer = TiktokenProxyTokenizer()

    def count_token(self, model_name: str, prompts: List[str]) -> List[int]:
        # Use tiktoken to count token in local environment
        return self._tiktoken_tokenizer.count_token(model_name, prompts)

    def support_async(self) -> bool:
        return True

    async def count_token_async(self, model_name: str, prompts: List[str]) -> List[int]:
        """Count token of given messages.

        This is relying on the claude beta API, which is not available for some users.
        """
        from dbgpt.util.chat_util import run_async_tasks

        tasks = []
        model_name = model_name or "claude-3-5-sonnet-20241022"
        for prompt in prompts:
            request = ModelRequest(
                model=model_name, messages=[{"role": "user", "content": prompt}]
            )
            tasks.append(
                self.client.beta.messages.count_tokens(
                    model=model_name,
                    messages=request.messages,
                )
            )
        results = await run_async_tasks(tasks, self.concurrency_limit)
        return results


register_proxy_model_adapter(
    ClaudeLLMClient,
    supported_models=[
        ModelMetadata(
            model=[
                "claude-3-5-sonnet-20241022",
                "claude-3-5-sonnet-latest",
                "claude-3-5-haiku-20241022",
                "claude-3-5-haiku-latest",
            ],
            context_length=200 * 1024,
            max_output_length=8 * 1024,
            description="Claude 3.5 by Anthropic",
            link="https://docs.anthropic.com/en/docs/about-claude/models#model-names",
            function_calling=True,
        ),
        ModelMetadata(
            model=[
                "claude-3-opus-20240229",
                "claude-3-opus-latest",
                "claude-3-sonnet-20240229",
                "claude-3-haiku-20240307",
            ],
            context_length=200 * 1024,
            max_output_length=4 * 1024,
            description="Claude 3 by Anthropic",
            link="https://docs.anthropic.com/en/docs/about-claude/models#model-names",
            function_calling=True,
        ),
    ],
)



"""TripletExtractor class."""

import asyncio
import logging
from abc import ABC, abstractmethod
from typing import List, Optional

from dbgpt.core import HumanPromptTemplate, LLMClient, ModelMessage, ModelRequest
from dbgpt.rag.transformer.base import ExtractorBase

logger = logging.getLogger(__name__)


class LLMExtractor(ExtractorBase, ABC):
    """LLMExtractor class."""

    def __init__(self, llm_client: LLMClient, model_name: str, prompt_template: str):
        """Initialize the LLMExtractor."""
        self._llm_client = llm_client
        self._model_name = model_name
        self._prompt_template = prompt_template

    async def extract(self, text: str, limit: Optional[int] = None) -> List:
        """Extract by LLM."""
        return await self._extract(text, None, limit)

    async def batch_extract(
        self,
        texts: List[str],
        batch_size: int = 1,
        limit: Optional[int] = None,
    ) -> List:
        """Batch extract by LLM."""
        if batch_size < 1:
            raise ValueError("batch_size >= 1")

        results = []

        for i in range(0, len(texts), batch_size):
            batch_texts = texts[i : i + batch_size]

            # Create tasks for current batch
            extraction_tasks = [
                self._extract(text, None, limit) for text in batch_texts
            ]

            # Execute batch concurrently and wait for all to complete
            batch_results = await asyncio.gather(*extraction_tasks)
            results.extend(batch_results)

        return results

    async def _extract(
        self, text: str, history: str = None, limit: Optional[int] = None
    ) -> List:
        """Inner extract by LLM."""
        # limit check
        if limit and limit < 1:
            raise ValueError("optional argument limit >= 1")

        template = HumanPromptTemplate.from_template(self._prompt_template)

        messages = (
            template.format_messages(text=text, history=history)
            if history is not None
            else template.format_messages(text=text)
        )

        # use default model if needed
        if not self._model_name:
            models = await self._llm_client.models()
            if not models:
                raise Exception("No models available")
            self._model_name = models[0].model
            logger.info(f"Using model {self._model_name} to extract")

        model_messages = ModelMessage.from_base_messages(messages)
        request = ModelRequest(model=self._model_name, messages=model_messages)
        response = await self._llm_client.generate(request=request)

        if not response.success:
            code = str(response.error_code)
            reason = response.text
            logger.error(f"request llm failed ({code}) {reason}")
            return []

        if response.has_text:
            return self._parse_response(response.text, limit)
        else:
            return []

    def truncate(self):
        """Do nothing by default."""

    def drop(self):
        """Do nothing by default."""

    @abstractmethod
    def _parse_response(self, text: str, limit: Optional[int] = None) -> List:
        """Parse llm response."""
